#!/usr/local/bin/python
# -*- coding: utf-8 -*-
"""
The language of positive RIF conditions determines what can appear as a body (the
 if-part) of a rule supported by the basic RIF logic. As explained in Section 
 Overview, RIF's Basic Logic Dialect corresponds to definite Horn rules, and the
  bodies of such rules are conjunctions of atomic formulas without negation.
"""
from rdflib import Variable, BNode, URIRef, Literal, Namespace,RDF,RDFS
from rdflib.Collection import Collection
from rdflib.Graph import ConjunctiveGraph,QuotedGraph,ReadOnlyGraphAggregate, Graph
from rdflib.syntax.NamespaceManager import NamespaceManager

OWL    = Namespace("http://www.w3.org/2002/07/owl#")

def buildUniTerm((s,p,o),newNss=None):
    return Uniterm(p,[s,o],newNss=newNss)

class QNameManager(object):
    def __init__(self,nsDict=None):
        self.nsDict = nsDict and nsDict or {}
        self.nsMgr = NamespaceManager(Graph())
        self.nsMgr.bind('owl','http://www.w3.org/2002/07/owl#')
        self.nsMgr.bind('math','http://www.w3.org/2000/10/swap/math#')
        
    def bind(self,prefix,namespace):
        self.nsMgr.bind(prefix,namespace)

class SetOperator(object):
    def repr(self,operator):
        return "%s( %s )"%(operator,' '.join([repr(i) for i in self.formulae]))
    def __len__(self):
        return len(self.formulae)

class Condition(object):
    """
    CONDITION   ::= CONJUNCTION | DISJUNCTION | EXISTENTIAL | ATOMIC
    """
    def __iter__(self):
        for f in self.formulae:
            yield f

class And(QNameManager,SetOperator,Condition):
    """
    CONJUNCTION ::= 'And' '(' CONDITION* ')'
    
    >>> And([Uniterm(RDF.type,[RDFS.comment,RDF.Property]),
    ...      Uniterm(RDF.type,[OWL.Class,RDFS.Class])])
    And( rdf:Property(rdfs:comment) rdfs:Class(owl:Class) )
    """
    def __init__(self,formulae=None):
        self.formulae = formulae and formulae or []
        QNameManager.__init__(self)
        
    def n3(self):
        """
        >>> And([Uniterm(RDF.type,[RDFS.comment,RDF.Property]),
        ...      Uniterm(RDF.type,[OWL.Class,RDFS.Class])]).n3()
        u'rdfs:comment a rdf:Property .\\n owl:Class a rdfs:Class'
        
        """
#        if not [term for term in self if not isinstance(term,Uniterm)]:
#            g= Graph(namespace_manager = self.nsMgr)
#            g.namespace_manager= self.nsMgr
#            [g.add(term.toRDFTuple()) for term in self]
#            return g.serialize(format='n3')
#        else:
        return u' .\n '.join([i.n3() for i in self])
        
    def __repr__(self):
        return self.repr('And')
    
class Or(QNameManager,SetOperator,Condition):
    """
    DISJUNCTION ::= 'Or' '(' CONDITION* ')'
    
    >>> Or([Uniterm(RDF.type,[RDFS.comment,RDF.Property]),
    ...      Uniterm(RDF.type,[OWL.Class,RDFS.Class])])
    Or( rdf:Property(rdfs:comment) rdfs:Class(owl:Class) )
    """
    def __init__(self,formulae=None):
        self.formulae = formulae and formulae or []
        QNameManager.__init__(self)
        
    def __repr__(self):
        return self.repr('Or')

class Exists(Condition):
    """
    EXISTENTIAL ::= 'Exists' Var+ '(' CONDITION ')'
    >>> Exists(formula=Or([Uniterm(RDF.type,[RDFS.comment,RDF.Property]),
    ...                    Uniterm(RDF.type,[OWL.Class,RDFS.Class])]),
    ...        declare=[Variable('X'),Variable('Y')])
    Exists ?X ?Y ( Or( rdf:Property(rdfs:comment) rdfs:Class(owl:Class) ) )
    """
    def __init__(self,formula=None,declare=None):
        self.formula = formula
        self.declare = declare and declare or []    
    def __iter__(self):
        for term in self.formula: 
            yield term
    def n3(self):
        """
        """
        return self.formula.n3()
        #return u"@forSome %s %s"%(','.join(self.declare),self.formula.n3())
    
    def __repr__(self):
        return "Exists %s ( %r )"%(' '.join([var.n3() for var in self.declare]),
                                   self.formula )
        
class Atomic(Condition):
    """
    ATOMIC ::= Uniterm | Equal | Member | Subclass (| Frame)
    """
    def __iter__(self):
        yield self

class Equal(QNameManager,Atomic):
    """
    Equal ::= TERM '=' TERM
    TERM ::= Const | Var | Uniterm | 'External' '(' Expr ')'
    
    >>> Equal(RDFS.Resource,OWL.Thing)
    rdfs:Resource =  owl:Thing
    """
    def __init__(self,lhs=None,rhs=None):
        self.lhs = lhs
        self.rhs = rhs
        QNameManager.__init__(self)
        
    def __repr__(self):
        left  = self.nsMgr.qname(self.lhs)
        right = self.nsMgr.qname(self.rhs)
        return "%s =  %s"%(left,right)

def BuildUnitermFromTuple((s,p,o),newNss=None):
    return Uniterm(p,[s,o],newNss)

class Uniterm(QNameManager,Atomic):
    """
    Uniterm ::= Const '(' TERM* ')'
    TERM ::= Const | Var | Uniterm
    
    We restrict to binary predicates (RDF triples)
    
    >>> Uniterm(RDF.type,[RDFS.comment,RDF.Property])
    rdf:Property(rdfs:comment)
    """
    def __init__(self,op,arg=None,newNss=None):        
        self.op = op
        self.arg = arg and arg or []
        QNameManager.__init__(self)
        if newNss is not None:
            newNss = isinstance(newNss,dict) and newNss.items() or newNss
            for k,v in newNss:
                self.nsMgr.bind(k,v)
        self._hash=hash(reduce(lambda x,y:str(x)+str(y),
            len(self.arg)==2 and self.toRDFTuple() or [self.op]+self.arg))
                
    def __hash__(self):
        return self._hash
    
    def __eq__(self,other):
        return hash(self) == hash(other)                       
        
    def renderTermAsN3(self,term):
        if term == RDF.type:
            return 'a'
        elif isinstance(term, (BNode,Literal,Variable)):
            return term.n3()
        else:
            return self.nsMgr.qname(term)
        
    def n3(self):
        """
        Serialize as N3 (using available namespace managers)
        
        >>> Uniterm(RDF.type,[RDFS.comment,RDF.Property]).n3()
        u'rdfs:comment a rdf:Property'

        """
        return ' '.join([ self.renderTermAsN3(term) 
                         for term in [self.arg[0],self.op,self.arg[1]]])
        
    def toRDFTuple(self):
        subject,_object = self.arg
        return (subject,self.op,_object)
        
    def collapseName(self,val):
        try:
            return self.nsMgr.qname(val)
        except:
            return val
        
    def normalizeTerm(self,term):
        if isinstance(term,Literal):
            return term.n3()
        else:
            return isinstance(term,Variable) and term.n3() or \
                   self.collapseName(term)        
        
    def __repr__(self):
        if self.op == RDF.type:
            arg0,arg1 = self.arg
            return "%s(%s)"%(self.normalizeTerm(arg1),self.normalizeTerm(arg0))
        else:
            return "%s(%s)"%(self.normalizeTerm(self.op),
                                ' '.join([self.normalizeTerm(i) for i in self.arg]))
            
class ExternalFunction(Uniterm):
    """
    An External(ATOMIC) is a call to an externally defined predicate, equality, 
    membership, subclassing, or frame. Likewise, External(Expr) is a call to an 
    externally defined function.
    >>> ExternalFunction(Uniterm(URIRef('http://www.w3.org/2000/10/swap/math#greaterThan'),[Variable('VAL'),Literal(2)]))
    math:greaterThan(?VAL 2)
    """
    def __init__(self,builtin,newNss=None):
        from FuXi.Rete.RuleStore import N3RuleStore,N3Builtin
        self.builtin= builtin
        if isinstance(builtin,N3Builtin):
            Uniterm.__init__(self,builtin.uri,[builtin.argument,builtin.result])
        else:
            Uniterm.__init__(self,builtin.op,builtin.arg)
        QNameManager.__init__(self)
        if newNss is not None:
            newNss = isinstance(newNss,dict) and newNss.items() or newNss
            for k,v in newNss:
                self.nsMgr.bind(k,v)
            
def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()


from pprint import pprint, pformat
from sets import Set
from FuXi.Rete import *
from FuXi.Syntax.InfixOWL import *
from FuXi.Rete.AlphaNode import SUBJECT,PREDICATE,OBJECT,VARIABLE
from FuXi.Rete.RuleStore import N3RuleStore
from FuXi.Rete.Util import renderNetwork,generateTokenSet
from FuXi.Horn.PositiveConditions import Uniterm, BuildUnitermFromTuple
from FuXi.Horn.HornRules import HornFromN3
from FuXi.DLP import MapDLPtoNetwork, non_DHL_OWL_Semantics
from FuXi.Rete.Magic import MagicSetTransformation, iterCondition, AdornLiteral
from rdflib.Namespace import Namespace
from rdflib import plugin,RDF,RDFS,URIRef,URIRef
from rdflib.OWL import FunctionalProperty
from rdflib.store import Store
from cStringIO import StringIO
from rdflib.Graph import Graph,ReadOnlyGraphAggregate,ConjunctiveGraph
from rdflib.syntax.NamespaceManager import NamespaceManager
from glob import glob
from rdflib.sparql.bison import Parse
import unittest, os, time, itertools

RDFLIB_CONNECTION=''
RDFLIB_STORE='IOMemory'

CWM_NS    = Namespace("http://cwmTest/")
DC_NS     = Namespace("http://purl.org/dc/elements/1.1/")
STRING_NS = Namespace("http://www.w3.org/2000/10/swap/string#")
MATH_NS   = Namespace("http://www.w3.org/2000/10/swap/math#")
FOAF_NS   = Namespace("http://xmlns.com/foaf/0.1/") 
OWL_NS    = Namespace("http://www.w3.org/2002/07/owl#")
TEST_NS   = Namespace("http://metacognition.info/FuXi/DL-SHIOF-test.n3#")
LOG       = Namespace("http://www.w3.org/2000/10/swap/log#")
RDF_TEST  = Namespace('http://www.w3.org/2000/10/rdf-tests/rdfcore/testSchema#')
OWL_TEST  = Namespace('http://www.w3.org/2002/03owlt/testOntology#')

MAGIC_PROOFS = True

queryNsMapping={'test':'http://metacognition.info/FuXi/test#',
                'rdf':'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                'foaf':'http://xmlns.com/foaf/0.1/',
                'dc':'http://purl.org/dc/elements/1.1/',
                'rss':'http://purl.org/rss/1.0/',
                'rdfs':'http://www.w3.org/2000/01/rdf-schema#',
                'rdf':'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                'owl':OWL_NS,
                'rdfs':RDF.RDFNS,
}

nsMap = {
  u'rdfs' :RDFS.RDFSNS,
  u'rdf'  :RDF.RDFNS,
  u'rete' :RETE_NS,
  u'owl'  :OWL_NS,
  u''     :TEST_NS,
  u'otest':OWL_TEST,
  u'rtest':RDF_TEST,
  u'foaf' :URIRef("http://xmlns.com/foaf/0.1/"),
  u'math' :URIRef("http://www.w3.org/2000/10/swap/math#"),
}

MANIFEST_QUERY = \
"""
SELECT ?status ?premise ?conclusion ?feature ?descr
WHERE {
  [ 
    a otest:PositiveEntailmentTest;
    otest:feature ?feature;
    rtest:description ?descr;
    rtest:status ?status;
    rtest:premiseDocument ?premise;
    rtest:conclusionDocument ?conclusion 
  ]
}"""
PARSED_MANIFEST_QUERY = Parse(MANIFEST_QUERY)

Features2Skip = [
    URIRef('http://www.w3.org/2002/07/owl#sameClassAs'),
]

Features2SkipMagic = [
    #FunctionalProperty
]

Tests2Skip = [
    #'OWL/oneOf/Manifest003.rdf', #logical set equivalence?  (works for magic sets)
    #'OWL/differentFrom/Manifest002.rdf'  ,#needs log:notEqualTo (works for magic sets)
    #'OWL/distinctMembers/Manifest001.rdf',#needs log:notEqualTo (works for magic sets)
    #'OWL/unionOf/Manifest002.rdf',#can't implement set theoretic union for owl:unionOf. (works for magic sets)
    'OWL/InverseFunctionalProperty/Manifest001.rdf',#owl:sameIndividualAs deprecated
    'OWL/FunctionalProperty/Manifest001.rdf', #owl:sameIndividualAs deprecated
    'OWL/Nothing/Manifest002.rdf',# owl:sameClassAs deprecated
    #'OWL/AllDifferent/Manifest001.rdf',#requires support for built-ins (log:notEqualTo), works for magic sets
    
    
    #Fix for magic set implementation
    #'OWL/TransitiveProperty/Manifest001.rdf', #recursive logic program
    #'OWL/InverseFunctionalProperty/Manifest002.rdf', #need to investigate
]

patterns2Skip = [
    'OWL/cardinality',
    'OWL/samePropertyAs'
]

class OwlTestSuite(unittest.TestCase):
    def setUp(self):
        store = plugin.get(RDFLIB_STORE,Store)()
        store.open(RDFLIB_CONNECTION)
        self.ruleStore=N3RuleStore()
        self.ruleGraph = Graph(self.ruleStore)
        self.ruleFactsGraph = Graph(store)
#        if not MAGIC_PROOFS:
#            self.ruleGraph.parse(StringIO(non_DHL_OWL_Semantics),format='n3')
        self.network = ReteNetwork(self.ruleStore,nsMap=nsBinds)
        
        #renderNetwork(self.network,nsMap=nsMap).write_graphviz('owl-rules.dot')
    def tearDown(self):
        pass
    
    def calculateEntailments(self, factGraph):
#        print self.network
#        print "Explicit (base) facts"
#        pprint([BuildUnitermFromTuple(i) for i in factGraph])
#        pprint(list(factGraph))
        start = time.time()  
        self.network.feedFactsToAdd(generateTokenSet(factGraph))                    
        sTime = time.time() - start
        if sTime > 1:
            sTimeStr = "%s seconds"%sTime
        else:
            sTime = sTime * 1000
            sTimeStr = "%s milli seconds"%sTime
        print "Time to calculate closure on working memory: ",sTimeStr
        print self.network
        
        tNodeOrder = [tNode 
                        for tNode in self.network.terminalNodes 
                            if self.network.instanciations.get(tNode,0)]
        tNodeOrder.sort(key=lambda x:self.network.instanciations[x],reverse=True)
        for termNode in tNodeOrder:
            print termNode
            print "\t", termNode.clause
            print "\t\t%s instanciations"%self.network.instanciations[termNode]
    #                    for c in AllClasses(factGraph):
    #                        print CastClass(c,factGraph)
        print "=============="
        self.network.inferredFacts.namespace_manager = factGraph.namespace_manager
#        if self.network.inferredFacts:
#            print "Implicit facts: "
#            print self.network.inferredFacts.serialize(format='turtle')
#        print "ruleset after MST:"                    
#        pprint(list(self.network.rules))
#        print "rate of reduction in the size of the program: ", len len(self.network.rules)
        return sTimeStr

    def MagicOWLProof(self,goals,rules,factGraph):
        print "Goals",[AdornLiteral(goal) for goal in goals]
        assert not self.network.rules
        progLen = len(rules)
        magicRuleNo = 0
        for rule in MagicSetTransformation(factGraph,
                                           rules,
                                           goals):
            magicRuleNo+=1
            self.network.buildNetworkFromClause(rule.formula)    
            self.network.rules.add(rule)
        print "rate of reduction in the size of the program: ", (100-(float(magicRuleNo)/float(progLen))*100)
        for goal in goals:
            factGraph.add(goal)
        timing=self.calculateEntailments(factGraph)
        for goal in goals:
            if goal not in self.network.inferredFacts and goal not in factGraph:
                print "missing triple %s"%(pformat(goal))
                raise #Exception ("Failed test: "+feature)
            else:
                print "=== Passed! ==="
        return timing
       
    def testOwl(self): 
        testData = {}       
        for manifest in glob('OWL/*/Manifest*.rdf'):
            if manifest in Tests2Skip:
                continue
            skip = False
            for pattern2Skip in patterns2Skip:
                if manifest.find(pattern2Skip) > -1:
                    skip = True
                    break
            if skip:
                continue            
            manifestStore = plugin.get(RDFLIB_STORE,Store)()
            manifestGraph = Graph(manifestStore)
            manifestGraph.parse(open(manifest))
            rt = manifestGraph.query(
                                      PARSED_MANIFEST_QUERY,
                                      initNs=nsMap,
                                      DEBUG = False)
            #print list(manifestGraph.namespace_manager.namespaces())
            for status,premise,conclusion, feature, description in rt:
                if feature in Features2Skip:
                    continue
                premise = manifestGraph.namespace_manager.compute_qname(premise)[-1]
                conclusion = manifestGraph.namespace_manager.compute_qname(conclusion)[-1]
                premiseFile    = '/'.join(manifest.split('/')[:2]+[premise])
                conclusionFile = '/'.join(manifest.split('/')[:2]+[conclusion])
                print premiseFile
                print conclusionFile
                if status == 'APPROVED':
                    assert os.path.exists('.'.join([premiseFile,'rdf'])) 
                    assert os.path.exists('.'.join([conclusionFile,'rdf']))
                    print "<%s> :- <%s>"%('.'.join([conclusionFile,'rdf']),
                                          '.'.join([premiseFile,'rdf']))
                    store = plugin.get(RDFLIB_STORE,Store)()
                    store.open(RDFLIB_CONNECTION)
                    factGraph = Graph(store)
                    factGraph.parse(open('.'.join([premiseFile,'rdf'])))
                    allFacts = ReadOnlyGraphAggregate([factGraph,self.ruleFactsGraph])
                    Individual.factoryGraph=factGraph
                    
                    for c in AllClasses(factGraph):
                        if not isinstance(c.identifier,BNode):
                            print c.__repr__(True)       
                            
                    if MAGIC_PROOFS:
                        if feature in Features2SkipMagic:
                            continue
                        print premiseFile,feature,description
                        program=list(HornFromN3(StringIO(non_DHL_OWL_Semantics)))
                        program.extend(self.network.setupDescriptionLogicProgramming(
                                                                     factGraph,
                                                                     addPDSemantics=False,
                                                                     constructNetwork=False))
#                        print "Original program"
#                        pprint(program)
                        timings=[]  
                        #Magic set algorithm is initiated by a query, a single horn ground, 'fact'                    
                        try:   
                            goals=set()
                            for triple in Graph(store).parse('.'.join([conclusionFile,'rdf'])):
                                if triple not in factGraph:
                                    goals.add(triple)
                            timings.append(self.MagicOWLProof(goals,program,factGraph))
                            self.network._resetinstanciationStats()
                            self.network.reset()
                            self.network.clear()                                
                        except:
                            print "missing triple %s"%(pformat(goal))
                            print manifest, premiseFile
                            print "feature: ", feature
                            print description
#                            from FuXi.Rete.Util import renderNetwork 
#                            pprint([BuildUnitermFromTuple(t) for t in self.network.inferredFacts])
#                            dot=renderNetwork(self.network,self.network.nsMap).write_jpeg('test-fail.jpeg')  
                            raise #Exception ("Failed test: "+feature)
                            
                        testData[manifest] = timings    
                        continue
                    else:
                        self.network.setupDescriptionLogicProgramming(factGraph,addPDSemantics=True)
                        sTimeStr=self.calculateEntailments(factGraph)
                        expectedFacts = Graph(store)                    
                        for triple in expectedFacts.parse('.'.join([conclusionFile,'rdf'])):
                            closureGraph = ReadOnlyGraphAggregate([self.network.inferredFacts,factGraph])
                            if triple not in self.network.inferredFacts and triple not in factGraph:
                                print "missing triple %s"%(pformat(triple))
                                print manifest
                                print "feature: ", feature
                                print description
                                pprint(list(self.network.inferredFacts))  
#                                pprint(programs)                          
                                raise Exception ("Failed test: "+feature)
                            else:
                                print "=== Passed! ==="
                                #pprint(list(self.network.inferredFacts))
                        print "\n"
                        testData[manifest] = sTimeStr
                        store.rollback()
                        self.network.reset()
                        self.network._resetinstanciationStats()
                    
#        pprint(testData)

def runTests(profile=False):
    suite = unittest.makeSuite(OwlTestSuite)
    if profile:
        #from profile import Profile
        from hotshot import Profile, stats
        p = Profile('fuxi.profile')
        #p = Profile()
        p.runcall(unittest.TextTestRunner(verbosity=5).run,suite)
        p.close()    
        s = stats.load('fuxi.profile')
#        s=p.create_stats()
        s.strip_dirs()
        s.sort_stats('time','cumulative','pcalls')
        s.print_stats(.1)
        s.print_callers(.05)
        s.print_callees(.05)
    else:
        unittest.TextTestRunner(verbosity=5).run(suite)
                
if __name__ == '__main__':
    runTests(profile=True)


"""
"""
from RuleStore import N3Builtin
from rdflib import Variable, BNode,RDF,Variable,Literal,RDFS, URIRef, Namespace
from rdflib.Graph import Graph 
from sets import Set
from ReteVocabulary import RETE_NS
from BuiltinPredicates import FILTERS, FUNCTIONS
from Node import Node

OWL_NS    = Namespace("http://www.w3.org/2002/07/owl#")

SUBJECT   = 0
PREDICATE = 1
OBJECT    = 2

VARIABLE = 0
VALUE    = 1

TERMS = [SUBJECT,PREDICATE,OBJECT]

def normalizeTerm(term):
    """
    Graph Identifiers are used
    """
    if isinstance(term,Graph):
        return term.identifier
    else:
        return term
        
from cPickle import dumps, PicklingError # for memoize
class memoize(object):
    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated. Slow for mutable types."""
    # Ideas from MemoizeMutable class of Recipe 52201 by Paul Moore and
    # from memoized decorator of http://wiki.python.org/moin/PythonDecoratorLibrary
    # For a version with timeout see Recipe 325905
    # For a self cleaning version see Recipe 440678
    # Weak references (a dict with weak values) can be used, like this:
    #   self._cache = weakref.WeakValueDictionary()
    #   but the keys of such dict can't be int
    def __init__(self, func):
        self.func = func
        self._cache = {}
    def __call__(self, *args, **kwds):
        key = args
        if kwds:
            items = kwds.items()
            items.sort()
            key = key + tuple(items)
        try:
            if key in self._cache:
                return self._cache[key]
            self._cache[key] = result = self.func(*args, **kwds)
            return result
        except TypeError:
            try:
                dump = dumps(key)
            except PicklingError:
                return self.func(*args, **kwds)
            else:
                if dump in self._cache:
                    return self._cache[dump]
                self._cache[dump] = result = self.func(*args, **kwds)
                return result

class ReteToken:
    """
    A ReteToken, an RDF triple in a Rete network.  Once it passes an alpha node test,
    if will have unification substitutions per variable
    """
    def __init__(self,(subject,predicate,object_),debug = False):
        self.debug = debug
        self.subject   = (None,normalizeTerm(subject))
        self.predicate = (None,normalizeTerm(predicate))
        self.object_   = (None,normalizeTerm(object_))
        self.bindingDict = {}
        self._termConcat = self.concatenateTerms([self.subject,self.predicate,self.object_])
        self.hash = hash(self._termConcat) 
        self.inferred = False

    def __hash__(self):
        """
        
        >>> token1 = ReteToken((RDFS.domain,RDFS.domain,RDFS.Class))
        >>> token2 = ReteToken((RDFS.domain,RDFS.domain,RDFS.Class))
        >>> token1 == token2
        True
        >>> token1 in Set([token2])
        True
        """
        return self.hash 

    @memoize
    def concatenateTerms(terms):
        return reduce(lambda x,y:str(x)+str(y),[term[VALUE] for term in terms])

    def __eq__(self,other):
        return hash(self) == hash(other)   

    def alphaNetworkHash(self,termHash):
        """
        We store pointers to all the system's alpha memories in a hash table, indexed
        according to the particular values being tested. Executing the alpha network then becomes a
        simple matter of doing eight hash table lookups:

        >>> aNode1 = AlphaNode((Variable('Z'),RDF.type,Variable('A')))
        >>> aNode2 = AlphaNode((Variable('X'),RDF.type,Variable('C')))
        >>> token = ReteToken((URIRef('urn:uuid:Boo'),RDF.type,URIRef('urn:uuid:Foo')))
        >>> token.alphaNetworkHash(aNode1.alphaNetworkHash())
        u'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'
                
        """
        triple = list(self.asTuple())
        termHash = list(termHash)
        return ''.join([triple[idx] for idx in TERMS if termHash[idx] == '1'])        
                
    def unboundCopy(self,noSubsequentDebug=False):
        if noSubsequentDebug:
            return ReteToken((self.subject[VALUE],self.predicate[VALUE],self.object_[VALUE]))
        else:
            return ReteToken((self.subject[VALUE],self.predicate[VALUE],self.object_[VALUE]),self.debug)
                
    def __repr__(self):        
        return "<ReteToken: %s>"%(
            ','.join(["%s->%s"%(var,val) for var,val in [self.subject,self.predicate,self.object_] if isinstance(var,(Variable,BNode))])
        )
            
    def asTuple(self):
        return (self.subject[VALUE],self.predicate[VALUE],self.object_[VALUE])
        
    def bindVariables(self,alphaNode):
        """
        This function, called when a token passes a node test, associates token terms with variables
        in the node test         
        """
        if isinstance(alphaNode,BuiltInAlphaNode):
            self.pattern = list(alphaNode.n3builtin)
            self.subject   = (alphaNode.n3builtin.argument,self.subject[VALUE])
            self.predicate = (alphaNode.n3builtin.uri,self.predicate[VALUE])
            self.object_   = (alphaNode.n3builtin.result,self.object_[VALUE])
            assert not self.bindingDict,self.bindingDict
            bindHashItems = []
            for var,val in [self.subject,self.predicate,self.object_]:
                if var and isinstance(var,(Variable,BNode)) and var not in self.bindingDict:
                    self.bindingDict[var] = val
                    bindHashItems.append(var + val)
                else:
                    bindHashItems.append(val)
            #self.bindingDict := { var1 -> val1, var2 -> val2, ..  }
            self.hash = hash(reduce(lambda x,y:x+y,bindHashItems))
            return self            
        else:
            self.pattern = alphaNode.triplePattern
            self.subject   = (alphaNode.triplePattern[SUBJECT],self.subject[VALUE])
            self.predicate = (alphaNode.triplePattern[PREDICATE],self.predicate[VALUE])
            self.object_   = (alphaNode.triplePattern[OBJECT],self.object_[VALUE])
            assert not self.bindingDict,self.bindingDict
            bindHashItems = []
            for var,val in [self.subject,self.predicate,self.object_]:
                if var and isinstance(var,(Variable,BNode)) and var not in self.bindingDict:
                    self.bindingDict[var] = val
                    bindHashItems.append(var + val)
                else:
                    bindHashItems.append(val)
            #self.bindingDict := { var1 -> val1, var2 -> val2, ..  }
            self.hash = hash(reduce(lambda x,y:x+y,bindHashItems))
            return self            
    
def defaultIntraElementTest(aReteToken,triplePattern):
    """
    'Standard' Charles Forgy intra element token pattern test.
    """
    tokenTerms = [aReteToken.subject[VALUE],aReteToken.predicate[VALUE],aReteToken.object_[VALUE]]
    varBindings = {}
    for idx in [SUBJECT,PREDICATE,OBJECT]:
        tokenTerm   = tokenTerms[idx]
        patternTerm = triplePattern[idx]
        if not isinstance(patternTerm,(Variable,BNode)) and tokenTerm != patternTerm:
            return False            
        elif patternTerm in varBindings and varBindings[patternTerm] != tokenTerm:
            return False
        elif patternTerm not in varBindings:
            varBindings[patternTerm] = tokenTerm
    return True

class AlphaNode(Node):
    """
    Basic Triple Pattern Pattern check
    """
    def __init__(self,triplePatternOrFunc):
        self.relinked = False
        self.name = BNode()
        self.triplePattern = triplePatternOrFunc
        self.descendentMemory = []
        self.descendentBetaNodes = Set()
        self.builtin = bool(FUNCTIONS.get(self.triplePattern[PREDICATE]) or FILTERS.get(self.triplePattern[PREDICATE]))
        self.universalTruths = []

    def alphaNetworkHash(self,groundTermHash=False):
        """
        Thus, given a WME w, to determine which alpha memories w should be added to, we need only check whether
        any of these eight possibilities is actually present in the system.  (Some might not be present, since 
        there might not be any alpha memory corresponding to that particular combination of tests and 's.)
        
        0 - Variable
        1 - Ground term        
        
        >>> aNode1 = AlphaNode((Variable('P'),RDF.type,OWL_NS.InverseFunctionalProperty))
        >>> aNode2 = AlphaNode((Variable('X'),Variable('P'),Variable('Z')))
        >>> aNode1.alphaNetworkHash()
        ('0', '1', '1')
        >>> aNode2.alphaNetworkHash()
        ('0', '0', '0')
        >>> aNode1.alphaNetworkHash(groundTermHash=True)
        u'http://www.w3.org/1999/02/22-rdf-syntax-ns#typehttp://www.w3.org/2002/07/owl#InverseFunctionalProperty'
        """
        if groundTermHash:
            return ''.join([term for term in self.triplePattern if not isinstance(term,(BNode,Variable))])
        else:
            return tuple([isinstance(term,(BNode,Variable)) and '0' or '1' for term in self.triplePattern])

    def checkDefaultRule(self,defaultRules):
        """
        Check to see if the inter element test associated with this Alpha node may match
        the given 'default' conflict set.  If so, update universalTruths with the
        default conflict set token list which if matched, means the intra element test automatically
        passes
        """
        pass

    def __repr__(self):
        return "<AlphaNode: %s. Feeds %s beta nodes>"%(repr(self.triplePattern),len(self.descendentBetaNodes))

    def activate(self,aReteToken):
        from BetaNode import PartialInstanciation, LEFT_MEMORY, RIGHT_MEMORY, LEFT_UNLINKING
        #print aReteToken.asTuple()
        #aReteToken.debug = True
        aReteToken.bindVariables(self)
        for memory in self.descendentMemory:
            singleToken = PartialInstanciation([aReteToken],consistentBindings=aReteToken.bindingDict.copy())
#            print memory
#            print self
#            print self.descendentMemory
            if memory.position == LEFT_MEMORY:
                memory.addToken(singleToken)
            else:
                memory.addToken(aReteToken)                
            if memory.successor.leftUnlinkedNodes and len(memory) == 1 and LEFT_UNLINKING:
                #Relink left memory of successor
#                from Util import renderNetwork
#                from md5 import md5
#                from datetime import datetime
#                import os                                
#                print "Re-linking %s"%(memory.successor)
#                print "Re-linking triggered from %s"%(repr(self))
                for node in memory.successor.leftUnlinkedNodes:
#                    print "\trelinking to ", node, " from ", memory.position
                    #aReteToken.debug = True
                    if node.unlinkedMemory is None:
                        assert len(node.descendentMemory) == 1,"%s %s %s"%(node,
                                                                        node.descendentMemory,
                                                                        memory.successor)                        
                        disconnectedMemory = list(node.descendentMemory)[0]                        
                        
                    else:
                        disconnectedMemory = node.unlinkedMemory
                        node.descendentMemory.append(disconnectedMemory)
                        node.unlinkedMemory = None
#                    if aReteToken.debug:
#                        print "\t reattached memory ",str(disconnectedMemory) 
                    memory.successor.memories[LEFT_MEMORY] = disconnectedMemory                    
                    node.descendentBetaNodes.add(memory.successor)
                    #print memory.successor.memories[LEFT_MEMORY]
                    memory.successor.propagate(RIGHT_MEMORY,aReteToken.debug,wme=aReteToken)                    
                    #node._activate(singleToken,aReteToken.debug)
                    #print "Activating re-linked node", node
                    #node.propagate(None,aReteToken.debug)
#                    if memory.position == LEFT_MEMORY:
#                        node.propagate(memory.position,aReteToken.debug,singleToken)
#                    else:
#                        node.propagate(memory.position,aReteToken.debug,wme=aReteToken)

#                if memory.successor.network:
#                    dtNow = datetime.now().isoformat()
#                    fName = dtNow.replace(':','-').replace('.','-')
#                    renderNetwork(memory.successor.network).write_graphviz(fName+'.dot')
#                    os.popen ('dot -Tsvg -o %s %s'%(fName+'.svg',fName+'.dot'),'r')
#                    os.remove(fName+'.dot')
#                    print fName
                
                #self.relinked = True
                memory.successor.leftUnlinkedNodes = Set()
            if aReteToken.debug:
                    print "Added %s to %s"%(aReteToken,memory.successor)
            if memory.successor.aPassThru or not memory.successor.checkNullActivation(memory.position):
                if aReteToken.debug:
                    print "Propagated from %s"%(self)
                    print aReteToken.asTuple()
                if memory.position == LEFT_MEMORY:
                    memory.successor.propagate(memory.position,aReteToken.debug,singleToken)
                else:
                    memory.successor.propagate(memory.position,aReteToken.debug,wme=aReteToken)
            else:
                if aReteToken.debug:
                    print "skipped null right activation of %s from %s"%(memory.successor,self)
                                
class BuiltInAlphaNode(AlphaNode):
    """
    An Alpha Node for Builtins which doesn't participate in intraElement tests
    """
    def __init__(self,n3builtin):
        self.name = BNode()
        self.n3builtin = n3builtin
        self.descendentMemory = []
        self.descendentBetaNodes = Set()
        self.universalTruths = [] 
    def __iter__(self):
        yield self.n3builtin.argument
        yield self.n3builtin.result       

    def alphaNetworkHash(self,groundTermHash=False):
        if groundTermHash:
            return ''.join([term for term in self.n3builtin if not isinstance(term,(BNode,Variable))])
        else:
            return tuple([isinstance(term,(BNode,Variable)) and '0' or '1' for term in self.n3builtin])

    def __repr__(self):
        return "<BuiltInAlphaNode %s(%s),%s : Feeds %s beta nodes>"%(self.n3builtin.func,
                                                                     self.n3builtin.argument,
                                                                     self.n3builtin.result,len(self.descendentBetaNodes))
    
    def intraElementTest(self,aReteToken):
        pass    
    
def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()    


import unittest, os, time, sys
from FuXi.Horn.PositiveConditions import And, Or, Uniterm, Condition, Atomic,SetOperator,Exists
from FuXi.Horn.HornRules import Clause
from rdflib import BNode, RDF, Namespace, Variable, RDFS, URIRef

SNO=Namespace('tag:info@ihtsdo.org,2007-07-31:SNOMED-CT#')

class NominalRangeTransformTest(unittest.TestCase):
    def setUp(self):
        pass
    
    def testRuleDuplicate(self):
        bN1 = BNode('plqJDAbW787')
        bN2 = BNode('plqJDAbW6934')
        head1 = And([Uniterm(RDF.type,[bN1,SNO.EntireHeart]),
                     Uniterm(SNO.partOf,[Variable('X'),bN1])])
        head2 = And([Uniterm(RDF.type,[bN2,SNO.EntireHeart]),
                     Uniterm(SNO.partOf,[Variable('X'),bN2])])
        commonBody = Uniterm(RDF.type,[Variable('X'),SNO.HeartPart])
        r1 = Clause(commonBody,Exists(head1,[bN1]))
        r2 = Clause(commonBody,Exists(head2,[bN2]))
        
        self.failUnlessEqual(r1,r2,"The BNodes should not count in syntactic equivalence!")
        
        bN1 = BNode('lSTVNgnt1363')
        bN2 = BNode('lSTVNgnt3553')
        head1 = And([Uniterm(RDF.type,[bN1,SNO.EntireBodyAsAWhole]),
                     Uniterm(SNO.partOf,[Variable('X'),bN1])])
        head2 = And([Uniterm(RDF.type,[bN2,SNO.EntireBodyAsAWhole]),
                     Uniterm(SNO.partOf,[Variable('X'),bN2])])
        commonBody = Uniterm(RDF.type,[Variable('X'),SNO.BodySystemStructure])
        r1 = Clause(commonBody,Exists(head1,[bN1]))
        r2 = Clause(commonBody,Exists(head2,[bN2]))

        self.failUnlessEqual(r1,r2,"The BNodes should not count in equivalence!")
        
        a=set([r1])
        self.failUnless(r1 in a,"The BNodes should not count in equivalence!")

if __name__ == "__main__":
    unittest.main()



PROGRAM2=\
"""
@prefix ex: <http://doi.acm.org/10.1145/28659.28689#>.
{ ?X ex:flat ?Y } => { ?X ex:sg ?Y }.
{ ?X ex:up ?Z1 . ?Z1 ex:sg ?Z2. ?Z2 ex:flat ?Z3. ?Z3 ex:sg ?Z4. ?Z4 ex:down ?Y } => { ?X ex:sg ?Y }.
"""

PROGRAM1=\
"""
@prefix : <http://example.com#>.
{ ?X par ?Y } => { ?X anc ?Y }
{ ?X par ?Z . ?Z  anc ?Y } => { ?X anc ?Y }
"""

PARTITION_LP_DB_PREDICATES=\
"""
@prefix ex: <http://doi.acm.org/10.1145/16856.16859#>.
ex:a ex:father ex:b.
ex:b ex:parent ex:c.
ex:b ex:grandfather ex:d.
{ ?X ex:father ?Z. ?X ex:parent ?Y } => { ?X ex:grandfather ?Y }.
"""

NON_LINEAR_MS_QUERY=\
"""
PREFIX ex: <http://doi.acm.org/10.1145/28659.28689#>
SELECT * WHERE { ex:john ex:sg ?X }
"""


#!/usr/bin/env python
"""
rdf.py - Trio RDF Concepts
Copyright 2007, Sean B. Palmer, inamidst.com
Licensed under the Eiffel Forum License 2.

Package: http://inamidst.com/sw/trio/
"""

import operator, itertools, unicodedata

def error(msg): 
   import sys
   print >> sys.stderr, msg
   sys.exit(1)

class Message(unicode): 
   def __new__(cls, name): 
      msg = 'Import Error: Missing %s.py from %s'
      uri = 'http://inamidst.com/sw/trio/'
      return unicode.__new__(cls, msg % (name, uri))

   __getattr__ = lambda self, *args: error(self)
   __setattr__ = lambda self, *args: error(self)

def optional(name): 
   # @@ make this give better error messages
   try: exec 'import ' + name
   except ImportError: 
      return Message(name)
   return eval(name)

try: from __init__ import capabilities
except ImportError: 
   capabilities = False

if capabilities is False: 
   capabilities = ['grddl', 'notation3', 'ntriples', 'rdfxml', 
     'serialise', 'srx', 'turtle', 'web']

for name in capabilities: 
   o = optional(name)
   exec name + ' = o'
del o

class Triple(object): 
   counter = itertools.count(1)

   def __init__(self, subject, predicate, object): 
      self.subject = subject
      self.predicate = predicate
      self.object = object

      assert isinstance(subject, Term)
      assert isinstance(predicate, Term)
      assert isinstance(object, Term)

      self.ID = Triple.counter.next()

   def __hash__(self): 
      return hash(self.subject) ^ hash(self.predicate) \
             ^ hash(self.object) ^ 0x31337

   def __eq__(self, other): 
      if not isinstance(other, Triple): 
         return False
      return ((self.subject == other.subject) and 
              (self.predicate == other.predicate) and 
              (self.object == other.object))

   def __ne__(self, other): 
      return (not self.__eq__(other))

   def __repr__(self): 
      s, p, o = self.subject, self.predicate, self.object
      return 'Triple(%r, %r, %r)' % (s, p, o)
# t = Triple

class Store(object): 
   def __init__(self): 
      # These map terms to a set of triple IDs
      self.subjects = {}
      self.predicates = {}
      self.objects = {}

      # This maps triple IDs to triples
      self.triples = {}

   def add(self, triple): 
      # Check for duplicates
      if isinstance(triple.subject, Constant) or \
         isinstance(triple.predicate, Constant) or \
         isinstance(triple.predicate, Constant): 
         for t in self.get(triple.subject, triple.predicate, triple.object): 
            return False

      self.subjects.setdefault(triple.subject, set()).add(triple.ID)
      self.predicates.setdefault(triple.predicate, set()).add(triple.ID)
      self.objects.setdefault(triple.object, set()).add(triple.ID)

      self.triples[triple.ID] = triple
      return True

   def get(self, subject=None, predicate=None, object=None): 
      if not (subject or predicate or object): 
         for triple in self.triples.itervalues(): 
            yield triple
         return

      sets = (db.get(term) or set() for (db, term) in 
              ((self.subjects, subject), 
               (self.predicates, predicate), 
               (self.objects, object)) if term)
      for tID in reduce(operator.__and__, sets): 
         yield self.triples[tID]

def resolve(relative): 
   """Resolve a relative URI."""
   import os, urlparse
   pwd = os.environ.get('PWD')
   if pwd is None: 
      pwd = os.getcwd()
   if not pwd.endswith('/'): 
      pwd += '/'
   return urlparse.urljoin('file://' + pwd, relative)

class Graph(object): 
   counter = itertools.count(1)
   DefaultStore = Store

   def __init__(self, uri=None, baseURI=None, format=None): 
      self.ID = Graph.counter.next()

      # Graph(uri=None, parse=True, format=None)
      # 
      # Graph() - won't parse
      # Graph(uri) - will parse
      # Graph(uri, parse=False) - won't parse
      # Graph(uri, parse=link) - will parse
      # 
      # Graph() - no parse
      # Graph(uri) - parse
      # Graph(base=uri) - no parse
      # Graph(uri, base=uri) - parse

      # if uri is None: 
      #    if parse is False: PASS
      #    elif parse is True: SET TO FALSE
      #    else: ERROR
      #    SET URI TO SOME FAKE
      # elif uri is not None: 
      #    if parse is False: 
      #       JUST SET THE URI
      #    elif parse is True: 
      #       SET THE URI, PARSE IT
      #    else: SET THE URI, BUT PARSE PARSE

      # if uri and parse: 
      #    if parse is not None: 
      #       source = parse
      #    else: source = uri
      # elif uri and (parse is False)

      self.uri = uri
      self.baseURI = baseURI or self.uri

      self.format = format
      self.store = Graph.DefaultStore()

      if self.uri: 
         self.parseURI(self.uri)

   def __len__(self): 
      return len(self.store.triples)

   def __eq__(self, other): 
      if not isinstance(other, Graph): 
         return False
      if len(self) != len(other): 
         return False

      for bindings in self.__query(other, bind=self.__bind_iso): 
         return True
      return False

   def __ne__(self, other): 
      return (not self.__eq__(other))

   def __iter__(self): 
      return self.store.triples.itervalues()

   def __add__(self, other): 
      G = Graph()
      for triple in self: 
         G.store.add(triple)
      for triple in other: 
         G.store.add(triple)
      return G

   def nodes(self): 
      for subject in self.store.subjects.iterkeys(): 
         yield subject
      for object in self.store.objects.iterkeys(): 
         if not self.store.subjects.has_key(object): 
            yield object

   def arcs(self): 
      return self.store.predicates.iterkeys()

   # def subjects(self): 
   #    return self.store.subjects.iterkeys()

   # def predicates(self): 
   #    return self.store.predicates.iterkeys()

   # def objects(self): 
   #    return self.store.objects.iterkeys()

   def uri(self): 
      return self.__uri

   def __setURI(self, uri): 
      if uri and (not ':' in uri): 
         uri = resolve(uri)
      self.__uri = uri
   uri = property(uri, __setURI)

   def baseURI(self): 
      return self.__baseURI

   def __setBaseURI(self, baseURI): 
      if baseURI and (not ':' in baseURI): 
         raise ValueError('baseURI must be absolute')
      elif not (self.uri or baseURI): # Make an API-instance unique URI
         baseURI = 'http://inamidst.com/misc/graphs/' + str(self.ID)
         # @@! use some function of the PWD instead, like cwm
      self.__baseURI = baseURI
   baseURI = property(baseURI, __setBaseURI)

   # @@ None of the parse* methods are to merge
   # Make a new graph and add that if you want to do that
   def parseText(self, text, format=None): 
      from cStringIO import StringIO
      f = StringIO(text.encode('utf-8'))
      f.seek(0)
      self.parseFile(f, format=format)
      f.close()

   def parseFile(self, f, format=None): 
      if format is None: 
         if self.format is None: 
            if isinstance(f, web.Document): 
               self.format = f.format()
            else: raise ValueError('A format is required for parsing')
         format = self.format

      if format == 'rdfxml': 
         rdfxml.URI = URI
         rdfxml.bNode = BlankNode
         rdfxml.Literal = PlainLiteral
         rdfxml.DatatypedLiteral = TypedLiteral

         def callback(s, p, o, self=self): 
            triple = Triple(s, p, o)
            self.store.add(triple)
         rdfxml.parseFile(self.baseURI, f, callback=callback)

      elif format == 'turtle': 
         turtle.URI = URI
         turtle.bNode = BlankNode
         turtle.Literal = PlainLiteral
         turtle.DatatypedLiteral = TypedLiteral

         class Parser(turtle.TurtleDocument): 
            def triple(self, s, p, o, graph=self): 
               triple = Triple(s, p, o)
               graph.store.add(triple)
         doc = Parser(self.baseURI, f)
         doc.parse()

      elif format == 'n3': 
         notation3.URI = URI
         notation3.bNode = BlankNode
         notation3.PlainLiteral = PlainLiteral
         notation3.TypedLiteral = TypedLiteral
         notation3.Var = Variable

         def triple(s, p, o, self=self): 
            self.store.add(Triple(s, p, o))
         sink = notation3.RDFSink(triple=triple)
         p = notation3.SinkParser(sink, baseURI=self.baseURI)
         p._bindings[''] = p._baseURI + '#'
         p.startDoc()
         bytes = f.read()
         p.feed(bytes)
         # for line in f: 
         #    p.feed(line)
         p.endDoc()

      elif format == 'ntriples': # @@ n-triples
         ntriples.URI = URI
         ntriples.bNode = BlankNode
         ntriples.PlainLiteral = PlainLiteral
         ntriples.TypedLiteral = TypedLiteral
         ntriples.Var = Variable
         ntriples.extended = True # @@ ntriples-rdf? n-triples! (?)

         class Parser(ntriples.Document): 
            def triple(self, s, p, o, graph=self): 
               triple = Triple(s, p, o)
               graph.store.add(triple)
         doc = Parser(self.baseURI, f)
         doc.parse()

      elif format == 'sparql-results': 
         srx.URI = URI
         srx.bNode = BlankNode
         srx.Literal = PlainLiteral
         srx.DatatypedLiteral = TypedLiteral

         class Parser(srx.Document): 
            def triple(self, s, p, o, graph=self): 
               triple = Triple(s, p, o)
               graph.store.add(triple)
         import xml.dom.minidom
         doc = xml.dom.minidom.parse(f)
         p = Parser(doc)
         p.parse()

      elif format == 'grddl': 
         for triple in grddl.parse(self.baseURI): # @@ urgh
            self.store.add(triple) # @@ urgh...
      else: raise ValueError('Unsupported format: %s' % format)

   def parseURI(self, uri, format=None): 
      u = web.doc(uri)
      self.parseFile(u, format=format)
      u.close()      

   # def addGraph(self, G): 
   #    pass

   def serialiseToLines(self, format):
      if format == 'rdfxml':
         return serialise.rdfxml(self)
      elif format == 'ntriples':
         return serialise.ntriples(self)
      raise Exception('%s is unsupported' % format)

   def serialiseToText(self, format): 
      lines = self.serialiseToLines(format)
      return u'\n'.join(lines)

   def serialiseToFile(self, format, f): 
      for line in self.serialiseToLines(format): 
         print >> f, line
      return True

   def varSorted(self): 
      import heapq
      heap = []

      def rank(term): 
         if isinstance(term, Variable): return 4
         if isinstance(term, BlankNode): return 3
         return 0

      for t in self: 
         ranking = rank(t.subject) + rank(t.predicate) + rank(t.object)
         heapq.heappush(heap, (ranking, t))

      while heap: 
         ranking, t = heapq.heappop(heap)
         yield t

   def __bind(self, (s, p, o), result_triple, bindings): 
      new_bindings = bindings.copy()

      if s is not None: 
         bound = new_bindings.get(s)
         if bound is None: 
            new_bindings[s] = result_triple.subject
         elif bound != result_triple.subject: 
            return False # Prune mismatched bindings

      if p is not None: 
         bound = new_bindings.get(p)
         if bound is None: 
            new_bindings[p] = result_triple.predicate
         elif bound != result_triple.predicate: 
            return False # Prune mismatched bindings

      if o is not None: 
         bound = new_bindings.get(o)
         if bound is None: 
            new_bindings[o] = result_triple.object
         elif bound != result_triple.object: 
            return False # Prune mismatched bindings

      return new_bindings

   def __bind_iso(self, (s, p, o), result_triple, bindings): 
      new_bindings = bindings.copy()

      if s is not None: 
         bound = new_bindings.get(s)
         if bound is None: 
            if type(s) != type(result_triple.subject): 
               return False # Prune mismatched types
            new_bindings[s] = result_triple.subject
         elif bound != result_triple.subject: 
            return False # Prune mismatched bindings

      if p is not None: 
         bound = new_bindings.get(p)
         if bound is None: 
            if type(p) != type(result_triple.predicate): 
               return False # Prune mismatched types
            new_bindings[p] = result_triple.predicate
         elif bound != result_triple.predicate: 
            return False # Prune mismatched bindings

      if o is not None: 
         bound = new_bindings.get(o)
         if bound is None: 
            if type(o) != type(result_triple.object): 
               return False # Prune mismatched types
            new_bindings[o] = result_triple.object
         elif bound != result_triple.object: 
            return False # Prune mismatched bindings

      return new_bindings

   def __query(self, Q, bind=None): 
      # Returns a generator of simple bindings
      if bind is None: 
         bind = self.__bind

      def masks(triple): 
         try: return masks.cache[triple]
         except KeyError: pass

         s = isinstance(triple.subject, Var)
         p = isinstance(triple.predicate, Var)
         o = isinstance(triple.object, Var)

         qmask = ((not s and triple.subject), 
                  (not p and triple.predicate), 
                  (not o and triple.object))
         bmask = ((s and triple.subject or None), 
                  (p and triple.predicate or None), 
                  (o and triple.object or None))

         masks.cache[triple] = qmask, bmask
         return qmask, bmask
      masks.cache = {}

      def product(inputs, bindings=None): 
         if bindings is None: 
            bindings = {}

         if inputs: 
            query_triple = inputs[0]
            qmask, bmask = masks(query_triple)
            result_triples = self.store.get(*qmask)
            for result_triple in result_triples: 
               new_bindings = bind(bmask, result_triple, bindings)
               if new_bindings is False: continue

               for result_bindings in product(inputs[1:], new_bindings):
                  yield result_bindings
         else: yield bindings

      query_triples = list(Q.varSorted())
      return product(query_triples)

   def query(self, Q, *args, **params): 
      sort, get = params.pop('sort', None), params.pop('get', None)

      if params: raise ValueError('Params must be sort and/or get')
      if args and (sort or get): 
         raise ValueError("Can't have args *and* params")

      if args: sort, get = args[0], args
      if get and (len(get) == 1): 
         get = get[0]

      # Create a generator of simple name: value query bindings
      results = (dict((k.label, v) for (k, v) in bindings.iteritems()
                      if isinstance(k, Variable)) # Variables only!
                 for bindings in self.__query(Q))

      if sort: 
         def compare(a, b): 
            return cmp(a[sort], b[sort])
         results = sorted(results, cmp=compare)

      if not get: 
         return results

      if isinstance(get, basestring): 
         return (b[get] for b in results)
      return (tuple(b[g] for g in get) for b in results)

   def subgraphs(self, Q): 
      # Query, returning subgraphs not bindings
      pass

   def get(self, subject=None, predicate=None, object=None): 
      print >> __import__('sys').stderr, """\
      Warning: Graph.get is deprecated! Use Graph.triples instead.
      """
      return self.store.get(subject, predicate, object)

   def triples(self, subject=None, predicate=None, object=None): 
      return self.store.get(subject, predicate, object)

   def the(self, subject=None, predicate=None, object=None): 
      result = None
      for triple in self.store.get(subject, predicate, object): 
         if result is None: 
            result = triple
         else: raise ValueError('Expected only one result, got more')
      return result

   def select(self, Q, order=None): 
      # @@ Remove this when people have had enough warning
      print >> __import__('sys').stderr, """\
      WARNING: Graph.select is deprecated!

         for bindings in G.select(Q, order='name'): 
            print bindings.name

      Becomes instead:

         for bindings in G.query(Q, sort='name'): 
            print bindings['name']

      Or better yet:

         for name in G.query(Q, 'name'): 
            print name
      """

      class Bindings(object): 
         pass

      results = []
      for bindings in self.__query(Q): 
         b = Bindings()
         for (k, v) in bindings.iteritems(): 
            if isinstance(k, Variable): 
               attr = k.label
               setattr(b, attr, v)

         if order is None: 
            yield b
         else: results.append(b)

      if order is not None: 
         def compare(a, b): 
            return cmp(getattr(a, order).lexical, getattr(b, order).lexical)
         for b in sorted(results, cmp=compare): 
            yield b

   def sparql(self, bgp): 
      """Evaluate a sparql.BGP instance against this Graph."""
      import sparql
      Q = Graph()
      for triple in bgp: 
         t = Triple(*triple)
         Q.store.add(t)
      results = self.__query(Q)

      class Multiset(sparql.Multiset): 
         def graph(self): 
            R = Graph()
            R.sparqlMultiset(self)
            if False: # @@ debug
               for line in R.serialiseToLines(format='ntriples'): 
                  print line
            return R

      multiset = Multiset()
      for bindings in results: 
         mapping = sparql.SolutionMapping(bindings)
         multiset.add(mapping)
      return multiset

   def sparqlMultiset(self, multiset): 
      rdf = u'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
      rs = u'http://www.w3.org/2001/sw/DataAccess/tests/result-set#'
      rdf_type = URI(rdf + u'type')
      rs_ResultSet = URI(rs + u'ResultSet')
      rs_resultVariable = URI(rs + u'resultVariable')
      rs_solution = URI(rs + u'solution')
      rs_binding = URI(rs + u'binding')
      rs_value = URI(rs + u'value')
      rs_variable = URI(rs + u'variable')      

      counter = itertools.count(1)
      s = BlankNode(u'set' + str(counter.next()))
      self.store.add(Triple(s, rdf_type, rs_ResultSet))

      def literal(var): 
         return PlainLiteral(unicode(var.label))

      for i, solution in enumerate(multiset): 
         if i == 0: 
            for variable in solution.iterkeys(): 
               self.store.add(Triple(s, rs_resultVariable, literal(variable)))
         sol = BlankNode(u'solution' + str(counter.next()))
         self.store.add(Triple(s, rs_solution, sol))
         for (variable, value) in solution.iteritems(): 
            bin = BlankNode(u'binding' + str(counter.next()))
            self.store.add(Triple(sol, rs_binding, bin))
            self.store.add(Triple(bin, rs_variable, literal(variable)))
            self.store.add(Triple(bin, rs_value, value))             

   def sparqlFile(self, rq, form=None): 
      """Evaluate a SPARQL file, rq, with this as the default Graph."""
      import sparql
      D = sparql.Dataset(default=self)
      # sparql.debug = True
      sparql.IRI = URI
      sparql.Variable = Variable
      E, DS, R = sparql.parse(rq.read())

      if form is not None: 
         if R[0] != form: 
            raise ValueError("Expected %s, got %s" % (form, R[0]))

      if R[0] == 'SELECT': 
         result = sparql.evaluate(D, E)
         vars = R[1]
         if vars and ((len(vars) > 1) or (vars[0] != '*')): 
            vars = set(vars)
            selected = type(result)()
            for solution in result: 
               sol = type(solution)()
               for variable, value in solution.iteritems(): 
                  if variable in vars: 
                     sol[variable] = value
               selected.add(sol)
            result = selected
         elif not vars: 
            result = sparql.Multiset()
      else: raise ValueError('Not implemented: %s' % R[0])

      if form is not None: 
         return result
      return R[0], result

   def sparqlURI(self, uri, form=None): 
      import urllib
      u = urllib.urlopen(uri)
      result = self.sparqlFile(u, form=form)
      u.close()
      return result

class Term(object): 
   """Any Constant (URI, Literal) or Var (BlankNode, Variable)."""

class Constant(Term): 
   """Any URI or Literal (PlainLiteral, TypedLiteral)."""

Monotonous = Constant

class URI(Constant): 
   # @@ relative vs. absolute
   def __init__(self, value): 
      assert isinstance(value, unicode)
      self.value = value

   def encode(self): 
      # or self.uri()
      raise Exception('@@ Not implemented')

   def validate(self): 
      # "does not contain any control characters (#x00-#x1F, #x7F-#x9F)"
      # check for control characters, check that it converts to URI okay
      raise Exception('@@ Not implemented')

   def __repr__(self): 
      return 'URI(%r)' % self.value

   def __str__(self): 
      """Return a NON-N-TRIPLES string representation."""
      return '<' + self.value.encode('utf-8') + '>'

   def __hash__(self): 
      return hash(self.value) ^ 0x31337

   def __eq__(self, other): 
      # @@ <class 'trio.rdf.URI'> <class 'rdf.URI'> False
      # if not isinstance(other, URI): 
      if not hasattr(other, 'value'): 
         return False
      return self.value == other.value

   def __ne__(self, other): 
      if not isinstance(other, URI): 
         return True
      return self.value != other.value

   def __cmp__(self, other): 
      if isinstance(other, URI): 
         return cmp(self.value, other.value)
      return -1      

class URIReference(object): 
   # @@ relative vs. absolute
   def __init__(self, value): 
      print >> __import__('sys').stderr, """
      WARNING: trio.rdf.URIReference is deprecated. Use trio.rdf.URI instead.
      """
      assert isinstance(value, unicode)
      self.value = value

   def encode(self): 
      # or self.uri()
      raise Exception('@@ Not implemented')

   def validate(self): 
      # "does not contain any control characters (#x00-#x1F, #x7F-#x9F)"
      # check for control characters, check that it converts to URI okay
      raise Exception('@@ Not implemented')

   def __repr__(self): 
      return 'URIReference(%r)' % self.value

   def __str__(self): 
      """Return a NON-N-TRIPLES string representation."""
      return '<' + self.value.encode('utf-8') + '>'

   def __hash__(self): 
      return hash(self.value) ^ 0x31337

   def __eq__(self, other): 
      # @@ <class 'trio.rdf.URIReference'> <class 'rdf.URIReference'> False
      # if not isinstance(other, URIReference): 
      if not hasattr(other, 'value'): 
         return False
      return self.value == other.value

   def __ne__(self, other): 
      if not isinstance(other, URIReference): 
         return True
      return self.value != other.value

   def __cmp__(self, other): 
      if isinstance(other, URIReference): 
         return cmp(self.value, other.value)
      return -1      

class Literal(Constant): 
   """An RDF Literal."""

# All literals have a lexical form being a Unicode [UNICODE] string, which 
# SHOULD be in Normal Form C [NFC].

class PlainLiteral(Literal): 
   """An RDF Plain Literal."""

   def __init__(self, lexical, language=None): 
      self.lexical = unicodedata.normalize('NFC', lexical)
      self.language = language

   def __hash__(self): 
      return hash(self.lexical) ^ hash(self.language) ^ 0x31337

   def __eq__(self, other): 
      if not isinstance(other, PlainLiteral): 
         return False
      if (self.lexical == other.lexical) and \
         (self.language == other.language): 
         return True
      return False

   def __ne__(self, other): 
      if not isinstance(other, PlainLiteral): 
         return True
      if (self.lexical == other.lexical) and \
         (self.language == other.language): 
         return False
      return True

   def __cmp__(self, other): 
      if isinstance(other, URI): 
         return 1
      elif isinstance(other, PlainLiteral): 
         return cmp((self.lexical, self.language), 
                    (other.lexical, other.language))
      return -1

   def __repr__(self): 
      return 'PlainLiteral(%r, %r)' % (self.lexical, self.language)

   def __str__(self): 
      """Return a NON-N-TRIPLES string representation."""
      s = '"' + self.lexical.encode('utf-8') + '"'
      if not self.language: 
         return s
      else: return s + '@' + self.language.encode('utf-8')

class TypedLiteral(Literal): 
   """An RDF Typed Literal."""

   def __init__(self, lexical, datatype): 
      self.lexical = unicodedata.normalize('NFC', lexical)
      assert isinstance(datatype, URI)
      self.datatype = datatype

   def __hash__(self): 
      return hash(self.lexical) ^ hash(self.datatype) ^ 0x31337

   def __eq__(self, other): 
      # This is *lexical* comparison, not value comparison
      if not isinstance(other, TypedLiteral): 
         return False
      if (self.lexical == other.lexical) and \
         (self.datatype == other.datatype): 
         return True
      return False

   def __ne__(self, other): 
      return (not self.__eq__(other))

   def __cmp__(self, other): 
      if isinstance(other, URI): 
         return 1
      elif isinstance(other, PlainLiteral): 
         return 1
      elif isinstance(other, TypedLiteral): 
         return cmp((self.lexical, self.datatype), 
                    (other.lexical, other.datatype))
      return -1

   def __repr__(self): 
      return 'TypedLiteral(%r, %r)' % (self.lexical, self.datatype)

   def __str__(self): 
      """Return a NON-N-TRIPLES string representation."""
      s = '"' + self.lexical.encode('utf-8') + '"'
      if not self.datatype: 
         return s
      else: return s + '^^' + str(self.datatype)

class Var(Term): 
   """Any BlankNode or Variable."""

Ticklish = Var

class BlankNode(Var): 
   def __init__(self, label): 
      self.label = label

   def __repr__(self): 
      return 'BlankNode(%r)' % self.label

   def __str__(self): 
      """Return a NON-N-TRIPLES string representation."""
      return '_:' + self.label.encode('utf-8')

   def __hash__(self): 
      return hash(self.label) ^ 0x31337

   def __eq__(self, other): 
      if not isinstance(other, BlankNode): 
         return False
      return self.label == other.label

   def __ne__(self, other): 
      if not isinstance(other, BlankNode): 
         return True
      return self.label != other.label

   def __cmp__(self, other): 
      if isinstance(other, URI): 
         return 1
      elif isinstance(other, PlainLiteral): 
         return 1
      elif isinstance(other, TypedLiteral): 
         return 1
      elif isinstance(other, BlankNode): 
         return cmp(self.label, other.label)
      return -1

class Variable(Var): 
   def __init__(self, label): 
      self.label = label

   def __repr__(self): 
      return 'Variable(%r)' % self.label

   def __hash__(self): 
      return hash(self.label) ^ 0x31337

   def __eq__(self, other): 
      if not isinstance(other, Variable): 
         return False
      return self.label == other.label

   def __ne__(self, other): 
      if not isinstance(other, Variable): 
         return True
      return self.label != other.label

   def __cmp__(self, other): 
      if isinstance(other, URI): 
         return 1
      elif isinstance(other, PlainLiteral): 
         return 1
      elif isinstance(other, TypedLiteral): 
         return 1
      elif isinstance(other, BlankNode): 
         return 1
      elif isinstance(other, Variable): 
         return cmp(self.label, other.label)
      return -1

class N3(object): 
   # Notation3 Environment
   def __init__(self): 
      # @@ IRIs
      self.bindings = {
         'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#', 
         'rdfs': 'http://www.w3.org/2000/01/rdf-schema#', 
         'owl': 'http://www.w3.org/2002/07/owl#', 
         'dc': 'http://purl.org/dc/elements/1.1/', 
         'foaf': 'http://xmlns.com/foaf/0.1/', 
         'doap': 'http://usefulinc.com/ns/doap#', 
         'skos': 'http://www.w3.org/2004/02/skos/core#'
      }

   def __call__(self, text, baseURI=None): 
      if not text.rstrip(' \t\r\n').endswith('.'): 
         text += ' .'
      prefixes = '\n'.join(
         '@prefix %s: <%s> .' % pair for pair in self.bindings.iteritems()
      ) + '\n@keywords a, is, of .\n\n'
      G = Graph(baseURI=baseURI)
      G.parseText(prefixes + text, format='n3')
      return G

   def __getitem__(self, qname): 
      if ':' in qname: 
         prefix, localname = qname.split(':')
      else: prefix, localname = '', qname
      namespace = self.bindings[prefix]
      return URI(unicode(namespace) + localname)

   def prefix(self, localname, uri): 
      self.bindings[localname] = uri

n3 = N3()

class Commands(object): 
   def __init__(self): 
      pass

   def format(self, uri): 
      resp = web.doc(uri)
      result = resp.format()
      resp.close()
      return result or 'unknown'

commands = Commands()

# EOF



import unittest, os, time, sys
from FuXi.Syntax.InfixOWL import *
from rdflib import plugin,RDF,RDFS,URIRef,URIRef,Literal,Variable
from rdflib.util import first
from rdflib.store import Store
from cStringIO import StringIO
from rdflib.Graph import Graph,ReadOnlyGraphAggregate,ConjunctiveGraph
from rdflib.syntax.NamespaceManager import NamespaceManager

def SubSumptionExpansion(owlClass):
    owlClass = CastClass(owlClass) 
    if isinstance(owlClass,BooleanClass) and owlClass._operator == OWL_NS.unionOf:
        for member in owlClass:
            expanded = False
            for innerMember in SubSumptionExpansion(Class(member)):
                expanded = True
                yield innerMember
            if not expanded:
                yield member
    else:
        for member in owlClass.subSumpteeIds():
            expanded = False
            for innerMember in SubSumptionExpansion(Class(member)):
                expanded = True
                yield innerMember
            if not expanded:
                yield member

def ComplementExpansion(owlClass,debug=False):
    owlClass=CastClass(owlClass.identifier,owlClass.graph)
    if isinstance(owlClass,BooleanClass) and \
       len(owlClass) == 2 and owlClass._operator == OWL_NS.intersectionOf:
        oldRepr = owlClass.__repr__()
        #A boolean-constructed class
        negativeClasses = set()
        otherClasses = set()
        for member in owlClass:
            member = Class(member)
            if member.complementOf:
                #A negative class, expand it and add to bucket of classes to 'remove'
                for expandedClass in SubSumptionExpansion(member.complementOf):
                    negativeClasses.add(expandedClass)
            else:
                #A positive class, expand it and add to bucket of base classes
                expanded = False
                for expandedClass in SubSumptionExpansion(member):
                    expanded = True
                    otherClasses.add(expandedClass)
                if not expanded:
                    otherClasses.add(member.identifier)

        if negativeClasses:
            #Delete the old list of operands for the boolean class
            oldList = owlClass._rdfList
            oldList.clear()
                        
            #Recreate the list of operands, exluding the expanded negative classes
            for allowedClasses in otherClasses.difference(negativeClasses) :
                oldList.append(classOrIdentifier(allowedClasses))
            owlClass.changeOperator(OWL_NS.unionOf)        
            if debug:
                print "Incoming boolean class: ", oldRepr             
                print "Expanded boolean class: ", owlClass.__repr__()
            return owlClass
        else:
            if debug:
                print "There were no negative classes!"
            
class ComplementExpansionTestSuite(unittest.TestCase):
    def setUp(self):
        self.testGraph = Graph()
        Individual.factoryGraph = self.testGraph         
    
    def testExpand(self):
        EX = Namespace("http://example.com/")
        namespace_manager = NamespaceManager(Graph())
        namespace_manager.bind('ex', EX, override=False)
        self.testGraph.namespace_manager = namespace_manager    
        
        man   = Class(EX.Man)
        boy   = Class(EX.Boy)
        woman = Class(EX.Woman)
        girl  = Class(EX.Girl)
        male  = Class(EX.Male)
        female= Class(EX.Female)
        human = Class(EX.Human)        
        animal = Class(EX.Animal)
        cat = Class(EX.Cat)
        dog = Class(EX.Dog)
        animal = Class(EX.Animal)
        
        animal = cat | dog | human
        human += man
        human += boy
        human += woman
        human += girl
        male   += man
        male   += boy
        female += woman
        female += girl
        
        testClass = human & ~ female
        self.assertEquals(repr(testClass),'( ex:Human and ( not ex:Female ) )')
        newtestClass = ComplementExpansion(testClass,debug=True)      
        self.assertTrue(repr(newtestClass) in ['( ex:Boy or ex:Man )','( ex:Man or ex:Boy )'],repr(newtestClass))

        testClass2 = animal & ~ (male | female)
        self.assertEquals(repr(testClass2),
                          '( ( ex:Cat or ex:Dog or ex:Human ) and ( not ( ex:Male or ex:Female ) ) )')
        newtestClass2 = ComplementExpansion(testClass2,debug=True)  
        testClass2Repr = repr(newtestClass2)
        self.assertTrue(testClass2Repr in ['( ex:Cat or ex:Dog )','( ex:Dog or ex:Cat )'],testClass2Repr)

if __name__ == '__main__':
    unittest.main()    
    sys.exit(1)
    from optparse import OptionParser
    parser = OptionParser()

    parser.add_option('--verbose',action="store_true",default=False,
        help='Output debug print statements or not')    
    parser.add_option('--format',default="xml",
        help='The RDF serialization syntax to parse with')

    (options, args) = parser.parse_args()

    owlGraph = Graph()
    for input in args[0:]:
        if options.verbose:
            print "Parsing ", input, " as ", options.format
        owlGraph.parse(input,format=options.format)

    Individual.factoryGraph = owlGraph
    
    def topList(node,g):
        for s in g.subjects(RDF.rest,node):
            yield s

    for negativeClass in owlGraph.subjects(predicate=OWL_NS.complementOf):
        containingList = first(owlGraph.subjects(RDF.first,negativeClass))
        prevLink = None
        while containingList:
            prevLink = containingList
            containingList = first(owlGraph.subjects(RDF.rest,containingList))
        for s,p,o in owlGraph.triples_choices((None,
                                            [OWL_NS.intersectionOf,
                                             OWL_NS.unionOf],
                                             prevLink)):
            _class = Class(s)
#            print _class.__repr__(True,True)            
            ComplementExpansion(_class,debug=options.verbose)


"""
See: http://www.w3.org/2000/10/swap/doc/CwmBuiltins
"""
import unittest, os, time, sys
from cStringIO import StringIO
from rdflib import Namespace, Variable, Literal, URIRef
from rdflib.Graph import Graph,ReadOnlyGraphAggregate,ConjunctiveGraph
from rdflib.util import first
STRING_NS = Namespace("http://www.w3.org/2000/10/swap/string#")
LOG_NS = Namespace("http://www.w3.org/2000/10/swap/log#")
MATH_NS = Namespace("http://www.w3.org/2000/10/swap/math#")
EULER_NS = Namespace("http://eulersharp.sourceforge.net/2003/03swap/owl-rules#")

def LogNotEqualTo(subject,object_):
    """
    Equality in this sense is actually the same URI.      
    """
    def func(s,o):
        return s != o
    return func

def LogEqualTo(subject,object_):
    """
    True if the subject and object are the same RDF node (symbol or literal).
    """
    def func(s,o):
        return s == o
    return func

def StringContains(subject,object_):
    return subject[-1].contains(object_[-1])

def StringGreaterThan(subject,object_):
    for term in [subject,object_]:
        if not isinstance(term,Variable):
            assert isinstance(term,Literal),"str:greaterThan can only be used with Literals! (%s)"%term
    def greaterThanF(s,o):
        for term in [s,o]:
            assert isinstance(term,Literal),"str:greaterThan can only be used with Literals!: %s %s"%(s,0)
        return str(s) > str(o)
    return greaterThanF

def StringLessThan(subject,object_):
    for term in [subject,object_]:
        if not isinstance(term,Variable):
            assert isinstance(term,Literal),"str:lessThan can only be used with Literals! (%s)"%term
    def lessThanF(s,o):
        for term in [s,o]:
            assert isinstance(term,Literal),"str:lessThan can only be used with Literals!"
        return s < o
        #return str(s) < str(o)
    return lessThanF

def StringEqualIgnoringCase(subject,object_):
    pass

#def MathProduct(arguments):
#    def productF(bindings):
#        return eval(' * '.join([isinstance(arg,Variable) and 'bindings[u"%s"]'%arg or str(arg) for arg in arguments]))
#    return productF
def MathEqualTo(subject,object_):
    for term in [subject,object_]:
        if not isinstance(term,Variable):
            assert isinstance(term,Literal),"math:equalTo can only be used with Literals! (%s)"%term
            assert isinstance(term.toPython(),(int,float,long)),"math:equalTo can only be used with Numeric Literals! (%s)"%term    
    def func(s,o):
        for term in [s,o]:
            assert isinstance(term,Literal),"math:equalTo can only be used with Literals!"
            assert isinstance(term.toPython(),(int,float,long)),"math:equalTo can only be used with Numeric Literals!"
        return s.toPython() == o.toPython()
    return func
def MathGreaterThan(subject,object_):
    for term in [subject,object_]:
        if not isinstance(term,Variable):
            assert isinstance(term,Literal),"math:lessThan can only be used with Literals! (%s)"%term
            assert isinstance(term.toPython(),(int,float,long)),"math:lessThan can only be used with Numeric Literals! (%s)"%term    
    def greaterThanF(s,o):
        for term in [s,o]:
            assert isinstance(term,Literal),"math:greaterThan can only be used with Literals!"
            assert isinstance(term.toPython(),(int,float,long)),"math:greaterThan can only be used with Numeric Literals!"
        return s.toPython() > o.toPython()
    return greaterThanF

def MathLessThan(subject,object_):
    for term in [subject,object_]:
        if not isinstance(term,Variable):
            assert isinstance(term,Literal),"math:lessThan can only be used with Literals! (%s)"%term
            assert isinstance(term.toPython(),(int,float,long)),"math:lessThan can only be used with Numeric Literals! (%s)"%term    
    def lessThanF(s,o):
        for term in [s,o]:
            assert isinstance(term,Literal),"math:lessThan can only be used with Literals!"
            assert isinstance(term.toPython(),(int,float,long)),"math:lessThan can only be used with Numeric Literals!"
        return s.toPython() < o.toPython()
    return lessThanF

def MathNotLessThan(subject,object_):
    for term in [subject,object_]:
        if not isinstance(term,Variable):
            assert isinstance(term,Literal),"math:notLessThan can only be used with Literals! (%s)"%term
            assert isinstance(term.toPython(),(int,float,long)),"math:lessThan can only be used with Numeric Literals! (%s)"%term    
    def nLessThanF(s,o):
        for term in [s,o]:
            assert isinstance(term,Literal),"math:notLessThan can only be used with Literals!"
            assert isinstance(term.toPython(),(int,float,long)),"math:lessThan can only be used with Numeric Literals!"
        return not(s.toPython() < o.toPython())
    return nLessThanF

FUNCTIONS = {
}
FILTERS = {
    LOG_NS.equalTo : LogEqualTo,
    LOG_NS.includes : None,
    LOG_NS.notEqualTo : LogNotEqualTo,
    LOG_NS.notIncludes : None,
    MATH_NS.equalTo : MathEqualTo,
    MATH_NS.greaterThan : MathGreaterThan,
    MATH_NS.lessThan : MathLessThan,
    MATH_NS.notEqualTo : None,
    MATH_NS.notGreaterThan : None,
    MATH_NS.notLessThan : MathNotLessThan,
    STRING_NS.contains : StringContains,
    STRING_NS.containsIgnoringCase : None,
    STRING_NS.endsWith : None,
    STRING_NS.equalIgnoringCase : StringEqualIgnoringCase,
    STRING_NS.greaterThan : StringGreaterThan,
    STRING_NS.lessThan : StringLessThan,
    STRING_NS.matches : None,
    STRING_NS.notEqualIgnoringCase : None,
    STRING_NS.notGreaterThan : None,
    STRING_NS.notLessThan : None,
    STRING_NS.notMatches : None,
    STRING_NS.startsWith : None,    
}

testN3="""
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix : <http://test/> .
@prefix math: <http://www.w3.org/2000/10/swap/math#> .
{ ?NODE rdf:value ?VAL. ?VAL math:greaterThan 2} => {?NODE :pred1 ?VAL}.
_:foo rdf:value 1;
      rdf:value 2;
      rdf:value 3.
"""

class NonEqualityPredicatesTestSuite(unittest.TestCase):
    def setUp(self):
        from FuXi.Rete.RuleStore import N3RuleStore
        from FuXi.Rete import ReteNetwork
        from FuXi.Rete.Util import generateTokenSet
        self.testGraph = Graph()
        self.ruleStore=N3RuleStore()
        self.ruleGraph = Graph(self.ruleStore)           
        self.ruleGraph.parse(StringIO(testN3),format='n3')
        self.testGraph.parse(StringIO(testN3),format='n3')        
        self.closureDeltaGraph = Graph()
        self.network = ReteNetwork(self.ruleStore,
                                   initialWorkingMemory=generateTokenSet(self.testGraph),
                                   inferredTarget = self.closureDeltaGraph,
                                   nsMap = {})
    def testParseBuiltIns(self):
        from FuXi.Rete.RuleStore import N3Builtin
        from FuXi.Rete.AlphaNode import BuiltInAlphaNode
        self.failUnless(self.ruleStore.rules>0, "No rules parsed out form N3!")
        for alphaNode in self.network.alphaNodes:
            if isinstance(alphaNode, BuiltInAlphaNode):
                self.failUnless(alphaNode.n3builtin.uri == MATH_NS.greaterThan, 
                                "Unable to find math:greaterThan func")

    def testEvaluateBuiltIns(self):
        from FuXi.Rete.RuleStore import N3Builtin
        from FuXi.Rete.AlphaNode import BuiltInAlphaNode
        self.failUnless(first(self.closureDeltaGraph.triples((None,URIRef('http://test/pred1'),Literal(3)))),
                            "Missing inferred :pred1 assertions")
    
if __name__ == '__main__':
    unittest.main()    



from pprint import pprint, pformat
from sets import Set
from FuXi.Rete import *
from FuXi.DLP import MapDLPtoNetwork, non_DHL_OWL_Semantics
from FuXi.Rete.AlphaNode import SUBJECT,PREDICATE,OBJECT,VARIABLE
from FuXi.Rete.Proof import GenerateProof
from FuXi.Rete.RuleStore import N3RuleStore
from FuXi.Horn.HornRules import Clause, Ruleset
from InfixOWL import Class
from FuXi.Rete.Util import renderNetwork,generateTokenSet
from FuXi.Horn.PositiveConditions import And, Or, Uniterm, Condition, Atomic,SetOperator, buildUniTerm
from rdflib.Namespace import Namespace
from rdflib import plugin,RDF,RDFS,URIRef,URIRef
from rdflib.store import Store
from cStringIO import StringIO
from rdflib.Graph import Graph,ReadOnlyGraphAggregate,ConjunctiveGraph
from rdflib.syntax.NamespaceManager import NamespaceManager
from glob import glob
from rdflib.sparql.bison import Parse
import unittest, os, time,sys

RDFLIB_CONNECTION=''
RDFLIB_STORE='IOMemory'

CWM_NS    = Namespace("http://cwmTest/")
DC_NS     = Namespace("http://purl.org/dc/elements/1.1/")
STRING_NS = Namespace("http://www.w3.org/2000/10/swap/string#")
MATH_NS   = Namespace("http://www.w3.org/2000/10/swap/math#")
FOAF_NS   = Namespace("http://xmlns.com/foaf/0.1/") 
OWL_NS    = Namespace("http://www.w3.org/2002/07/owl#")
TEST_NS   = Namespace("http://metacognition.info/FuXi/DL-SHIOF-test.n3#")
LOG       = Namespace("http://www.w3.org/2000/10/swap/log#")
RDF_TEST  = Namespace('http://www.w3.org/2000/10/rdf-tests/rdfcore/testSchema#')
OWL_TEST  = Namespace('http://www.w3.org/2002/03owlt/testOntology#')

queryNsMapping={'test':'http://metacognition.info/FuXi/test#',
                'rdf':'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                'foaf':'http://xmlns.com/foaf/0.1/',
                'dc':'http://purl.org/dc/elements/1.1/',
                'rss':'http://purl.org/rss/1.0/',
                'rdfs':'http://www.w3.org/2000/01/rdf-schema#',
                'rdf':'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                'owl':OWL_NS,
                'rdfs':RDF.RDFNS,
}

WITCH = Namespace('http://www.w3.org/2000/10/swap/test/reason/witch#')
DAN=Namespace('http://www.w3.org/2000/10/swap/test/reason/dan_home#')

nsMap = {
  u'rdfs' :RDFS.RDFSNS,
  u'rdf'  :RDF.RDFNS,
  u'rete' :RETE_NS,
  u'owl'  :OWL_NS,
  u''     :TEST_NS,
  u'otest':OWL_TEST,
  u'rtest':RDF_TEST,
  u'witch':WITCH,
  u'foaf' :URIRef("http://xmlns.com/foaf/0.1/"),
  u'math' :URIRef("http://www.w3.org/2000/10/swap/math#"),
}

class TestProof(unittest.TestCase):
    
    def createRuleStore(self,ruleGraphPath):
        ruleStore=N3RuleStore()
        Graph(ruleStore).parse(ruleGraphPath,format='n3')
        return ruleStore
    
    def setUp(self):
        store = plugin.get(RDFLIB_STORE,Store)()
        store.open(RDFLIB_CONNECTION)
        self.nsMapping = [('witch',WITCH),('dan',DAN)]
                
    def testMetacognition(self):
        g=Graph()
        g.parse('http://purl.org/net/chimezie/foaf')
        g.parse('http://xmlns.com/foaf/spec/')
        goal=(URIRef('http://flickr.com/photos/45452910@N00/21945461/'),
              RDF.type,
              FOAF_NS.Image)
        ruleStore=self.createRuleStore(StringIO(non_DHL_OWL_Semantics))
        network = ReteNetwork(ruleStore,
                              goal=goal,
                              nsMap=nsMap)
        MapDLPtoNetwork(network,g)
        factualTokens=generateTokenSet(g)
        start = time.time()  
        network.feedFactsToAdd(factualTokens)
        sTime = time.time() - start
        if sTime > 1:
            sTimeStr = "%s seconds"%sTime
        else:
            sTime = sTime * 1000
            sTimeStr = "%s milli seconds"%sTime
        print "Calculated closure in %s "%sTimeStr
        print network
        builder,proof=GenerateProof(network,goal)
        pGraph=Graph()
        builder.serialize(proof,pGraph)
        #print pGraph.serialize(format='n3')
#        for l in builder.trace:
#            print l

    def testProofWitch(self,
                  proof_viz="witches-pf",
                  network_viz='witches_network',
                  goal=(WITCH.GIRL,RDF.type,WITCH.WITCH)):
        ruleStore=self.createRuleStore('witch.n3')
        ruleStore._finalize()
        rs=Ruleset(n3StoreSrc=ruleStore.rules,nsMapping=self.nsMapping)
        network = ReteNetwork(ruleStore,
                              dontFinalize=True,
                              goal=goal,
                              nsMap=self.nsMapping)
        #renderNetwork(network,nsMap=dict(self.nsMapping)).write_graphviz('%s.dot'%network_viz)
        start = time.time()  
        factualTokens=generateTokenSet(Graph().parse('witch.n3',format='n3'))
        self.assertEqual(len(factualTokens),3)
        network.feedFactsToAdd(factualTokens)
        sTime = time.time() - start
        if sTime > 1:
            sTimeStr = "%s seconds"%sTime
        else:
            sTime = sTime * 1000
            sTimeStr = "%s milli seconds"%sTime
        print "Time to calculate closure on working memory: ",sTimeStr        
        print network
        self.failUnless(goal in network.inferredFacts,
                     "Didn't infer %s"%repr(goal))
        builder,proof=GenerateProof(network,goal)
        #builder.renderProof(proof).write_graphviz('%s.dot'%proof_viz)

    def testProofDan(self,proof_src='dan_home.n3',goal=(DAN.dan,DAN.homeRegion,DAN.Texas)):
        ruleStore=self.createRuleStore('dan_home.n3')
        ruleStore._finalize()
        rs=Ruleset(n3StoreSrc=ruleStore.rules,nsMapping=self.nsMapping)
        network = ReteNetwork(ruleStore,
                              dontFinalize=True,
                              goal=goal,
                              nsMap=self.nsMapping)
        start = time.time()  
        factualTokens=generateTokenSet(Graph().parse(proof_src,format='n3'))
        self.assertEqual(len(factualTokens),2,repr([buildUniTerm(i.asTuple(),network.nsMap) for i in factualTokens]))
        network.feedFactsToAdd(factualTokens)
        sTime = time.time() - start
        if sTime > 1:
            sTimeStr = "%s seconds"%sTime
        else:
            sTime = sTime * 1000
            sTimeStr = "%s milli seconds"%sTime
        print "Time to calculate closure on working memory: ",sTimeStr        
        print network
        self.failUnless(goal in network.inferredFacts,
                     "Didn't infer %s"%repr(goal))
        builder,proof=GenerateProof(network,goal)
        
def runTests(profile=False):
    suite = unittest.makeSuite(TestProof)
    if profile:
        #from profile import Profile
        from hotshot import Profile, stats
        p = Profile('fuxi.profile')
        #p = Profile()
        p.runcall(unittest.TextTestRunner(verbosity=5).run,suite)
        p.close()    
        s = stats.load('fuxi.profile')
#        s=p.create_stats()
        s.strip_dirs()
        s.sort_stats('time','cumulative','pcalls')
        s.print_stats(.1)
        s.print_callers(.05)
        s.print_callees(.05)
    else:
        unittest.TextTestRunner(verbosity=5).run(suite)
                
if __name__ == '__main__':
    runTests(profile=False)


"""
Implements the behavior associated with the 'join' (Beta) node in a RETE network:
    - Stores tokens in two memories
    - Tokens in memories are checked for consistent bindings (unification) for variables in common *across* both
    - Network 'trigger' is propagated downward
    
This reference implementation follows,  quite closely, the algorithms presented in the PhD thesis (1995) of Robert Doorenbos:
    Production Matching for Large Learning Systems (RETE/UL)
    
A N3 Triple is a working memory element (WME)

The Memories are implemented with consistent binding hashes. Unlinking is not implemented but null 
activations are mitigated (somewhat) by the hash / Set mechanism.
              
"""
import unittest, os, time, sys
from itertools import izip, ifilter
from pprint import pprint
from AlphaNode import AlphaNode, BuiltInAlphaNode, ReteToken
from Node import Node
from RuleStore import N3Builtin
from IteratorAlgebra import hash_join
from Util import xcombine
from rdflib import Variable, BNode,RDF,RDFS,Literal
from rdflib.Collection import Collection
from sets import Set
from itertools import izip
from ReteVocabulary import RETE_NS
from rdflib.Graph import QuotedGraph, Graph
from rdflib.Literal import _XSD_NS
from rdflib import BNode, RDF, Namespace, URIRef, Literal, Variable
OWL_NS    = Namespace("http://www.w3.org/2002/07/owl#")
Any = None

LEFT_MEMORY  = 1
RIGHT_MEMORY = 2

#Implementn left unlinking?
LEFT_UNLINKING = False

memoryPosition = {
    LEFT_MEMORY : "left",
    RIGHT_MEMORY: "right",
}

def collectVariables(node):
    """
    Utility function for locating variables common to the patterns in both left and right nodes
    """
    if isinstance(node,BuiltInAlphaNode):
        return Set()
    if isinstance(node,AlphaNode):
        return Set([term for term in node.triplePattern if isinstance(term,(Variable,BNode))])
    elif node:
        combinedVars = Set()
        combinedVars |= node.leftVariables
        combinedVars |= node.rightVariables
        return combinedVars
    else:
        return Set()
        
#From itertools recipes
def iteritems(mapping): 
    return izip(mapping.iterkeys(),mapping.itervalues())
    
def any(seq,pred=None):
    """Returns True if pred(x) is true for at least one element in the iterable"""
    for elem in ifilter(pred,seq):
        return True
    return False

class ReteMemory(Set):
    """
    A hashed rete network memory
    
    The hash setup goes like so:
    
    key = (commonVarValues1,commonVarValues2, .. commonVarValuesN)
    
    where commonVarValuesN are the values associated with the variables common to both the left
    and right side the descendant beta (join) node.  The unique values (variable substitutions)
    are mapped to a set of the corresponding tokens.  These substitutions are 'joined' on in order
    to create partial instanciations of tokens from both sides.   
    
    Memory mapping:
    {
      key => [ .. tripleOrToken ..]
    }   
    
    "The hash function is a function of both the appropriate variable binding from the token or WME
    .. and the node itself."     

    ".. One way to deal with this problem would be to have beta memory nodes check for duplicate
    tokens. Every time a beta memory was activated, it would check whether the \new" match
    was actually a duplicate of some token already in the memory; if so, it would be ignored (i.e.,
    discarded)."
    
    Memories are Set mixins..    

    procedure beta-memory-left-activation (node: beta-memory, t: token, w: WME)
        new-token   allocate-memory()
        new-token.parent <- t
        new-token.wme  <- w
        insert new-token at the head of node.items
        for each child in node.children do left-activation (child, new-token)
    end        

    procedure alpha-memory-activation (node: alpha-memory, w: WME)
        insert w at the head of node.items
        for each child in node.successors do right-activation (child, w)
    end    
     
    """
    def __init__(self,betaNode,position,filter=None):
        super(ReteMemory, self).__init__()
        self.filter                = filter
        self.successor             = betaNode
        self.position              = position
        self.substitutionDict      = {} #hashed 
        
    def union(self, other):
        """Return the union of two sets as a new set.

        (I.e. all elements that are in either set.)
        """
        result = ReteMemory(self.successor,self.position)
        result._update(other)
        return result    
            
    def __repr__(self):
        return "<%sMemory: %s item(s)>"%(self.position == LEFT_MEMORY and 'Beta' or 'Alpha', len(self))

    def addToken(self,token,debug=False):
        """       

        >>> aNode1 = AlphaNode((Variable('P'),RDF.type,OWL_NS.InverseFunctionalProperty))
        >>> aNode2 = AlphaNode((Variable('X'),Variable('P'),Variable('Z')))
        >>> aNode3 = AlphaNode((Variable('Y'),Variable('P'),Variable('Z')))
        >>> token1 = ReteToken((URIRef('urn:uuid:bart'),URIRef('urn:uuid:name'),Literal("Bart Simpson")))
        >>> token1 = token1.bindVariables(aNode2)
        >>> token2 = token1.unboundCopy()
        >>> token2 = token2.bindVariables(aNode3)
        >>> token3 = ReteToken((URIRef('urn:uuid:b'),URIRef('urn:uuid:name'),Literal("Bart Simpson")))
        >>> token3 = token3.bindVariables(aNode2)
        >>> token4 = token3.unboundCopy()
        >>> token4 = token4.bindVariables(aNode3)
        >>> token5 = ReteToken((URIRef('urn:uuid:name'),RDF.type,OWL_NS.InverseFunctionalProperty))
        >>> token5 = token5.bindVariables(aNode1)
        >>> joinNode1 = BetaNode(aNode1,aNode2)
        >>> joinNode2 = BetaNode(joinNode1,aNode3)
        >>> joinNode1.connectIncomingNodes(aNode1,aNode2)        
        >>> cVars = { Variable('P') : URIRef('urn:uuid:name'), Variable('Z') : Literal("Bart Simpson") }
        >>> inst = PartialInstanciation([token1,token2,token3,token4,token5],consistentBindings=cVars)
        >>> inst
        <PartialInstanciation (joined on ?P,?Z): Set([<ReteToken: Y->urn:uuid:b,P->urn:uuid:name,Z->Bart Simpson>, <ReteToken: P->urn:uuid:name>, <ReteToken: X->urn:uuid:b,P->urn:uuid:name,Z->Bart Simpson>, <ReteToken: Y->urn:uuid:bart,P->urn:uuid:name,Z->Bart Simpson>, <ReteToken: X->urn:uuid:bart,P->urn:uuid:name,Z->Bart Simpson>])>
        >>> b = ReteMemory(joinNode2,LEFT_MEMORY)
        >>> b.addToken(inst)
        >>> b
        <BetaMemory: 1 item(s)>
        >>> joinNode2
        <BetaNode : CommonVariables: [u'P', u'Z'] (0 in left, 0 in right memories)>
        >>> b.substitutionDict.keys()
        [(u'urn:uuid:name', rdflib.Literal('Bart Simpson',language=None,datatype=None))]
        """
        commonVarKey = []
        if isinstance(token,PartialInstanciation):
            for binding in token.bindings:
                commonVarKey = []
                for var in self.successor.commonVariables:
                    commonVarKey.append(binding.get(var))
                self.substitutionDict.setdefault(tuple(commonVarKey),Set()).add(token)
        else:
            for var in self.successor.commonVariables:            
                commonVarKey.append(token.bindingDict.get(var))        
            self.substitutionDict.setdefault(tuple(commonVarKey),Set()).add(token)
        self.add(token)
                            
    def reset(self):
        self.clear()
        self.substitutionDict = {}

def project(orig_dict, attributes,inverse=False):
    """
    Dictionary projection: http://jtauber.com/blog/2005/11/17/relational_python:_projection
    
    >>> a = {'one' : 1, 'two' : 2, 'three' : 3 }
    >>> project(a,['one','two'])
    {'two': 2, 'one': 1}
    >>> project(a,['four'])
    {}
    >>> project(a,['one','two'],True)
    {'three': 3}
    """
    if inverse:
        return dict([item for item in orig_dict.items() if item[0] not in attributes])
    else:
        return dict([item for item in orig_dict.items() if item[0] in attributes])
        
class PartialInstanciation(object):
    """
    Represents a set of WMEs 'joined' along one or more
    common variables from an ancestral join node 'up' the network
    
    In the RETE/UL PhD thesis, this is refered to as a token, which contains a set of WME triples.
    This is a bit of a clash with the use of the same word (in the original Forgy paper) to 
    describe what is essentially a WME and whether or not it is an addition to the networks memories 
    or a removal
    
    It is implemented (in the RETE/UL thesis) as a linked list of:
        
    structure token:
        parent: token {points to the higher token, for items 1...i-1} 
        wme: WME {gives item i}
    end
    
    Here it is instead implemented as a Set of WME triples associated with a list variables whose
    bindings are consistent
    
    
    >>> aNode = AlphaNode((Variable('X'),RDF.type,Variable('C')))
    >>> token = ReteToken((URIRef('urn:uuid:Boo'),RDF.type,URIRef('urn:uuid:Foo')))
    >>> token = token.bindVariables(aNode)
    >>> PartialInstanciation([token])    
    <PartialInstanciation: Set([<ReteToken: X->urn:uuid:Boo,C->urn:uuid:Foo>])>
    >>> for token in PartialInstanciation([token]):
    ...   print token
    <ReteToken: X->urn:uuid:Boo,C->urn:uuid:Foo>
    """
    def __init__(self,tokens = None,debug = False,consistentBindings = None):
        """
        Note a hash is calculated by 
        sorting & concatenating the hashes of its tokens 
        """
        self.joinedBindings = consistentBindings and consistentBindings or {}
        self.inconsistentVars = Set()
        self.debug = debug
        self.tokens = Set()
        self.bindings = []
        if tokens:
            for token in tokens:
                self.add(token,noPostProcessing=True)        
            self._generateHash()
            self._generateBindings()

    def _generateHash(self):
        tokenHashes = [hash(token) for token in self.tokens]
        tokenHashes.sort()
        self.hash = hash(reduce(lambda x,y:x+y,tokenHashes))        

    def unify(self,left,right):
        """
        Takes two dictionary and collapses it if there are no overlapping 'bindings' or
        'rounds out' both dictionaries so they each have each other's non-overlapping binding 
        """
        bothKeys = [key for key in left.keys() + right.keys() if key not in self.joinedBindings]
        if len(bothKeys) == len(Set(bothKeys)):
            joinDict = left.copy()
            joinDict.update(right)
            return joinDict
        else:
            rCopy = right.copy()
            left.update(project(rCopy,[key for key in right.keys() if key not in left]))          
            lCopy = left.copy()
            right.update(project(lCopy,[key for key in left.keys() if key not in right]))
            return [left,right]

    def _generateBindings(self):
        """
        Generates a list of dictionaries - each a unique variable substitution (binding)
        which applies to the ReteTokens in this PartialInstanciation
    
        >>> aNode = AlphaNode((Variable('S'),Variable('P'),Variable('O')))
        >>> token1 = ReteToken((URIRef('urn:uuid:alpha'),OWL_NS.differentFrom,URIRef('urn:uuid:beta')))
        >>> token2 = ReteToken((URIRef('urn:uuid:beta'),OWL_NS.differentFrom,URIRef('urn:uuid:alpha')))
        >>> cVars = { Variable('P') : OWL_NS.differentFrom }
        >>> inst = PartialInstanciation([token1.bindVariables(aNode),token2.bindVariables(aNode)],consistentBindings=cVars)
        >>> inst
        <PartialInstanciation (joined on ?P): Set([<ReteToken: S->urn:uuid:beta,P->http://www.w3.org/2002/07/owl#differentFrom,O->urn:uuid:alpha>, <ReteToken: S->urn:uuid:alpha,P->http://www.w3.org/2002/07/owl#differentFrom,O->urn:uuid:beta>])>
        >>> inst.joinedBindings
        {u'P': u'http://www.w3.org/2002/07/owl#differentFrom'}
        >>> inst.tokens
        Set([<ReteToken: S->urn:uuid:beta,P->http://www.w3.org/2002/07/owl#differentFrom,O->urn:uuid:alpha>, <ReteToken: S->urn:uuid:alpha,P->http://www.w3.org/2002/07/owl#differentFrom,O->urn:uuid:beta>])
        >>> inst.bindings        
        [{u'P': u'http://www.w3.org/2002/07/owl#differentFrom', u'S': u'urn:uuid:beta', u'O': u'urn:uuid:alpha'}, {u'P': u'http://www.w3.org/2002/07/owl#differentFrom', u'S': u'urn:uuid:alpha', u'O': u'urn:uuid:beta'}]
        
        Ensure unjoined variables with different names aren't bound to the same value 
        (B and Y aren't both bound to "Bart Simpson" simultaneously)
        
        >>> aNode1 = AlphaNode((Variable('A'),URIRef('urn:uuid:name'),Variable('B')))
        >>> aNode2 = AlphaNode((Variable('X'),URIRef('urn:uuid:name'),Variable('Y')))
        >>> token1 = ReteToken((URIRef('urn:uuid:bart'),URIRef('urn:uuid:name'),Literal("Bart Simpson")))
        >>> token1 = token1.bindVariables(aNode1)
        >>> token2 = ReteToken((URIRef('urn:uuid:b'),URIRef('urn:uuid:name'),Literal("Bart Simpson")))
        >>> token2 = token2.bindVariables(aNode2)
        >>> inst = PartialInstanciation([token1,token2])
        >>> pprint(inst.bindings)
        [{u'A': u'urn:uuid:bart',
          u'B': rdflib.Literal('Bart Simpson',language=None,datatype=None),
          u'X': u'urn:uuid:b',
          u'Y': rdflib.Literal('Bart Simpson',language=None,datatype=None)}]

        Ensure different variables which bind to the same value *within* a token includes this combination
        in the resulting bindings

        >>> aNode1 = AlphaNode((Variable('P1'),RDF.type,URIRef('urn:uuid:Prop1')))
        >>> aNode2 = AlphaNode((Variable('P2'),RDF.type,URIRef('urn:uuid:Prop1')))
        >>> aNode3 = AlphaNode((Variable('P1'),Variable('P2'),RDFS.Class))
        >>> token1 = ReteToken((RDFS.domain,RDFS.domain,RDFS.Class))
        >>> token2 = ReteToken((RDFS.domain,RDF.type,URIRef('urn:uuid:Prop1')))
        >>> token3 = ReteToken((RDFS.range,RDF.type,URIRef('urn:uuid:Prop1')))
        >>> token4 = ReteToken((RDFS.range,RDFS.domain,RDFS.Class))
        >>> inst = PartialInstanciation([token1.bindVariables(aNode3),token2.bindVariables(aNode1),token3.bindVariables(aNode2),token4.bindVariables(aNode3)])
        >>> len(inst.bindings)
        3
        >>> inst.bindings
        [{u'P2': u'http://www.w3.org/2000/01/rdf-schema#range', u'P1': u'http://www.w3.org/2000/01/rdf-schema#domain'}, {u'P2': u'http://www.w3.org/2000/01/rdf-schema#domain', u'P1': u'http://www.w3.org/2000/01/rdf-schema#domain'}, {u'P2': u'http://www.w3.org/2000/01/rdf-schema#domain', u'P1': u'http://www.w3.org/2000/01/rdf-schema#range'}]
                
        >>> aNode1 = AlphaNode((Variable('X'),RDF.value,Literal(2)))
        >>> aNode2 = AlphaNode((Variable('X'),RDF.type,Variable('Y')))                            
        >>> aNode3 = AlphaNode((Variable('Z'),URIRef('urn:uuid:Prop1'),Variable('W')))
        >>> token2 = ReteToken((URIRef('urn:uuid:Foo'),RDF.value,Literal(2))).bindVariables(aNode1)
        >>> token3 = ReteToken((URIRef('urn:uuid:Foo'),RDF.type,URIRef('urn:uuid:Baz'))).bindVariables(aNode2)
        >>> token5 = ReteToken((URIRef('urn:uuid:Bar'),URIRef('urn:uuid:Prop1'),URIRef('urn:uuid:Beezle'))).bindVariables(aNode3)
        >>> inst = PartialInstanciation([token2,token3,token5],consistentBindings={Variable('X'):URIRef('urn:uuid:Foo')})
        >>> pprint(list(inst.tokens))
        [<ReteToken: Z->urn:uuid:Bar,W->urn:uuid:Beezle>,
         <ReteToken: X->urn:uuid:Foo>,
         <ReteToken: X->urn:uuid:Foo,Y->urn:uuid:Baz>]
        >>> inst.bindings
        [{u'Y': u'urn:uuid:Baz', u'X': u'urn:uuid:Foo', u'Z': u'urn:uuid:Bar', u'W': u'urn:uuid:Beezle'}]
        
        >>> inst = PartialInstanciation([token2],consistentBindings={Variable('X'):URIRef('urn:uuid:Foo')})
        >>> inst.bindings
        [{u'X': u'urn:uuid:Foo'}]
        
        >>> aNode1 = AlphaNode((Variable('P'),OWL_NS.inverseOf,Variable('Q')))
        >>> aNode2 = AlphaNode((Variable('P'),RDF.type,OWL_NS.InverseFunctionalProperty))
        >>> token1 = ReteToken((URIRef('urn:uuid:Foo'),OWL_NS.inverseOf,URIRef('urn:uuid:Bar'))).bindVariables(aNode1)
        >>> token2 = ReteToken((URIRef('urn:uuid:Foo'),RDF.type,OWL_NS.InverseFunctionalProperty)).bindVariables(aNode1)
        >>> inst = PartialInstanciation([token1,token2],consistentBindings={Variable('P'):URIRef('urn:uuid:Foo'),Variable('Q'):URIRef('urn:uuid:Bar')})
        >>> inst._generateBindings()
        >>> inst.bindings
        [{u'Q': u'urn:uuid:Bar', u'P': u'urn:uuid:Foo'}]
        """
        if len(self.tokens) == 1:
            self.bindings = [list(self.tokens)[0].bindingDict.copy()]
            return
        bindings  = []
        forcedBindings = []
        isolatedBindings = {}
        for token in self.tokens:
            noIterations = 0
            newDict = {}
            for key in ifilter(
                    lambda x:x not in self.joinedBindings,
                    token.bindingDict.keys()):            
                var = key
                newDict[var] = token.bindingDict[var]
                noIterations+=1
            if noIterations == 1:
                isolatedBindings.setdefault(var,Set()).add(token.bindingDict[var])
            elif noIterations > 1:
                forcedBindings.append(newDict)
        revIsolBindings = {}
        for vals in isolatedBindings.itervalues():
            for val in vals:
                revIsolBindings.setdefault(val,Set()).add(var)
        if isolatedBindings:
            for i in xcombine(*tuple([tuple([(key,val) for val in vals]) 
                        for key,vals in iteritems(isolatedBindings) ])):
                isolatedDict = dict(i)                
                for val in isolatedDict.itervalues():
                    keysForVal = revIsolBindings[val] 
                    if len(keysForVal) <= 1:
                        newDict = isolatedDict.copy()
                        newDict.update(self.joinedBindings)
                        if newDict not in bindings:
                            bindings.append(newDict)
        def collapse(left,right):
            if isinstance(left,list):
                if not left:
                    if isinstance(right,list):
                        return right
                    else:
                        return [right]
                elif isinstance(right,list):
                    return reduce([left,right])
                elif len(left)==1:
                    u = self.unify(left[0],right)
                    if isinstance(u,list):
                        return u 
                    else:
                        return [u]
                else:
                    return left+[right]
            elif isinstance(right,list) and not right and left:
                return [left]
            return self.unify(left,right)
        for forcedBinding in forcedBindings:
            newDict = forcedBinding.copy()
            newDict.update(self.joinedBindings)
            if newDict not in bindings:
                bindings.append(newDict)
        self.bindings = reduce(collapse,bindings,[])
        if not self.bindings:
            self.bindings = [self.joinedBindings]

    def __hash__(self):
        return self.hash 

    def __eq__(self,other):
        return hash(self) == hash(other)                   

    def add(self,token,noPostProcessing=False):
        """        
        >>> aNode = AlphaNode((Variable('S'),Variable('P'),Variable('O')))
        >>> token1 = ReteToken((URIRef('urn:uuid:Boo'),RDF.type,URIRef('urn:uuid:Foo')))
        >>> token2 = ReteToken((URIRef('urn:uuid:Foo'),RDF.type,URIRef('urn:uuid:Boo')))
        >>> inst = PartialInstanciation([token1.bindVariables(aNode),token2.bindVariables(aNode)])
        >>> inst    
        <PartialInstanciation: Set([<ReteToken: S->urn:uuid:Boo,P->http://www.w3.org/1999/02/22-rdf-syntax-ns#type,O->urn:uuid:Foo>, <ReteToken: S->urn:uuid:Foo,P->http://www.w3.org/1999/02/22-rdf-syntax-ns#type,O->urn:uuid:Boo>])>
        """
        self.tokens.add(token)        
        if not noPostProcessing:
            self._generateHash()
            self._generateBindings()
    
    def __repr__(self):
        if self.joinedBindings:
            joinMsg = ' (joined on %s)'%(','.join(['?'+v for v in self.joinedBindings]))
        else:
            joinMsg = ''
        return "<PartialInstanciation%s: %s>"%(joinMsg,self.tokens)

    def __iter__(self):
        return self.tokens.__iter__()
    
    def __len(self):
        return len(self.tokens)

    def addConsistentBinding(self,newJoinVariables):
        #newJoinDict = self.joinedBindings.copy()
        #only a subset of the tokens in this partial instanciation will be 'merged' with
        #the new token - joined on the new join variables
        newJoinDict = dict([(v,None) for v in newJoinVariables])
        #newJoinDict.update(dict([(v,None) for v in newJoinVariables]))
        for binding in self.bindings:
            for key,val in newJoinDict.iteritems():
                boundVal = binding.get(key)
                if boundVal is not None:
                    assert val is None or boundVal == val
                    if val is None:
                        newJoinDict[key]=boundVal
        self.joinedBindings.update(newJoinDict)
        self._generateBindings()             
        
    def newJoin(self,rightWME,newJoinVariables):
        """
        >>> aNode1 = AlphaNode((Variable('P1'),RDF.type,URIRef('urn:uuid:Prop1')))
        >>> aNode2 = AlphaNode((Variable('P2'),RDF.type,URIRef('urn:uuid:Prop1')))
        >>> aNode3 = AlphaNode((Variable('P1'),Variable('P2'),RDFS.Class))
        >>> token1 = ReteToken((RDFS.domain,RDFS.domain,RDFS.Class))
        >>> token2 = ReteToken((RDFS.domain,RDF.type,URIRef('urn:uuid:Prop1')))
        >>> token3 = ReteToken((RDFS.range,RDF.type,URIRef('urn:uuid:Prop1')))
        >>> token4 = ReteToken((RDFS.range,RDFS.domain,RDFS.Class))
        >>> token5 = ReteToken((RDFS.domain,RDF.type,URIRef('urn:uuid:Prop1'))).bindVariables(aNode2)
        >>> inst = PartialInstanciation([token2.bindVariables(aNode1),token3.bindVariables(aNode2),token5])
        >>> pprint(list(inst.tokens))
        [<ReteToken: P2->http://www.w3.org/2000/01/rdf-schema#range>,
         <ReteToken: P1->http://www.w3.org/2000/01/rdf-schema#domain>,
         <ReteToken: P2->http://www.w3.org/2000/01/rdf-schema#domain>]
        >>> newInst = inst.newJoin(token1.bindVariables(aNode3),[Variable('P2')])
        >>> token1
        <ReteToken: P1->http://www.w3.org/2000/01/rdf-schema#domain,P2->http://www.w3.org/2000/01/rdf-schema#domain>
        >>> newInst
        <PartialInstanciation (joined on ?P2): Set([<ReteToken: P1->http://www.w3.org/2000/01/rdf-schema#domain,P2->http://www.w3.org/2000/01/rdf-schema#domain>, <ReteToken: P1->http://www.w3.org/2000/01/rdf-schema#domain>, <ReteToken: P2->http://www.w3.org/2000/01/rdf-schema#domain>])>
        >>> pprint(list(newInst.tokens))
        [<ReteToken: P1->http://www.w3.org/2000/01/rdf-schema#domain,P2->http://www.w3.org/2000/01/rdf-schema#domain>,
         <ReteToken: P1->http://www.w3.org/2000/01/rdf-schema#domain>,
         <ReteToken: P2->http://www.w3.org/2000/01/rdf-schema#domain>]
        """
        newJoinDict = self.joinedBindings.copy()
        if newJoinVariables:
            #only a subset of the tokens in this partial instanciation will be 'merged' with
            #the new token - joined on the new join variables
            newJoinDict.update(project(rightWME.bindingDict,newJoinVariables))
            newPInst = PartialInstanciation([],consistentBindings=newJoinDict)
            for token in self.tokens:
                commonVars = False
                for newVar in ifilter(
                    lambda x:x in token.bindingDict and rightWME.bindingDict[x] == token.bindingDict[x],
                    newJoinVariables):
                    #consistent token
                    commonVars = True
                    newPInst.add(token,noPostProcessing=True)
                if not commonVars:
                    #there are no common variables, no need to check
                    newPInst.add(token,noPostProcessing=True)
        else:
            #all of the tokens in this partial instanciation are already bound consistently with
            #respect to the new token
            newPInst = PartialInstanciation([],consistentBindings=newJoinDict)
            for token in self.tokens:
                newPInst.add(token,noPostProcessing=True)
        newPInst.add(rightWME)
        return newPInst            
            
class BetaNode(Node):      
    """
    Performs a rete network join between partial instanciations in its left memory and tokens in its memories

    "The data structure for a join node, therefore, must contain pointers to its two memory
    nodes (so they can be searched), a specification of any variable binding consistency tests to be
    performed, and a list of the node's children. .. (the beta memory is always its parent)."
    
    Setup 3 alpha nodes (Triple Patterns):
        
        aNode1 = ?X rdf:value 1
        aNode2 = ?X rdf:type ?Y
        aNode3 = ?Z <urn:uuid:Prop1> ?W
    
    >>> aNode1 = AlphaNode((Variable('X'),RDF.value,Literal(2)))
    >>> aNode2 = AlphaNode((Variable('X'),RDF.type,Variable('Y')))                            
    >>> aNode3 = AlphaNode((Variable('Z'),URIRef('urn:uuid:Prop1'),Variable('W')))
    

    Rete Network
    ------------

   aNode1 
     |
  joinNode1
       \   aNode2  
        \   /    aNode3
       joinNode2  / 
           \     /
            \   /
         joinNode3    
        
    joinNode3 is the Terminal node
    
    >>> joinNode1 = BetaNode(None,aNode1,aPassThru=True)
    >>> joinNode1.connectIncomingNodes(None,aNode1)
    >>> joinNode2 = BetaNode(joinNode1,aNode2)
    >>> joinNode2.connectIncomingNodes(joinNode1,aNode2)    
    >>> joinNode3 = BetaNode(joinNode2,aNode3)
    >>> joinNode3.connectIncomingNodes(joinNode2,aNode3)

    >>> joinNode1
    <BetaNode (pass-thru): CommonVariables: [?X] (0 in left, 0 in right memories)>
    >>> joinNode2
    <BetaNode : CommonVariables: [?X] (0 in left, 0 in right memories)>

    Setup tokens (RDF triples):
        
        token1 = <urn:uuid:Boo> rdf:value 2
        token2 = <urn:uuid:Foo> rdf:value 2
        token3 = <urn:uuid:Foo> rdf:type <urn:uuid:Baz>             (fires network)
         
        token3 is set with a debug 'trace' so its path through the network is printed along the way
        
        token4 = <urn:uuid:Bash> rdf:type <urn:uuid:Baz>            
        token5 = <urn:uuid:Bar> <urn:uuid:Prop1> <urn:uuid:Beezle>  (fires network)
        token6 = <urn:uuid:Bar> <urn:uuid:Prop1> <urn:uuid:Bundle>  (fires network)

    >>> token1 = ReteToken((URIRef('urn:uuid:Boo'),RDF.value,Literal(2)))
    >>> token2 = ReteToken((URIRef('urn:uuid:Foo'),RDF.value,Literal(2)))
    >>> token3 = ReteToken((URIRef('urn:uuid:Foo'),RDF.type,URIRef('urn:uuid:Baz')),debug=True)
    >>> token4 = ReteToken((URIRef('urn:uuid:Bash'),RDF.type,URIRef('urn:uuid:Baz')))
    >>> token5 = ReteToken((URIRef('urn:uuid:Bar'),URIRef('urn:uuid:Prop1'),URIRef('urn:uuid:Beezle')),debug=True)
    >>> token6 = ReteToken((URIRef('urn:uuid:Bar'),URIRef('urn:uuid:Prop1'),URIRef('urn:uuid:Bundle')))
    >>> tokenList = [token1,token2,token3,token4,token5,token6]

    Setup the consequent (RHS) of the rule:
        { ?X rdf:value 1. ?X rdf:type ?Y. ?Z <urn:uuid:Prop1> ?W } => { ?X a <urn:uuid:SelectedVar> }
        
    a Network 'stub' is setup to capture the conflict set at the time the rule is fired
    
    >>> joinNode3.consequent.update([(Variable('X'),RDF.type,URIRef('urn:uuid:SelectedVar'))])
    >>> class NetworkStub:
    ...     def __init__(self):
    ...         self.firings = 0
    ...         self.conflictSet = Set()
    ...     def fireConsequent(self,tokens,termNode,debug):
    ...         self.firings += 1
    ...         self.conflictSet.add(tokens)
    >>> testHelper = NetworkStub()
    >>> joinNode3.network = testHelper

    Add the tokens sequentially to the top of the network (the alpha nodes).
    token3 triggers a trace through it's path down to the terminal node (joinNode2) 

    >>> aNode1.descendentMemory[0]
    <AlphaMemory: 0 item(s)>
    >>> aNode1.descendentMemory[0].position
    2
    >>> aNode1.activate(token1.unboundCopy())
    >>> aNode1.activate(token2.unboundCopy())
\    >>> joinNode1.memories[LEFT_MEMORY]
    <BetaMemory: 0 item(s)>
    >>> joinNode2.memories[LEFT_MEMORY]
    <BetaMemory: 2 item(s)>
    >>> aNode1.activate(token3.unboundCopy())
    Propagated from <AlphaNode: (u'X', u'http://www.w3.org/1999/02/22-rdf-syntax-ns#type', u'Y'). Feeds 1 beta nodes>
    (u'urn:uuid:Foo', u'http://www.w3.org/1999/02/22-rdf-syntax-ns#type', u'urn:uuid:Baz')
    <BetaNode : CommonVariables: [u'X'] (2 in left, 1 in right memories)>.propagate(right,None,<ReteToken: X->urn:uuid:Foo,Y->urn:uuid:Baz>)
    activating with <PartialInstanciation (joined on ?X): Set([<ReteToken: X->urn:uuid:Foo>, <ReteToken: X->urn:uuid:Foo,Y->urn:uuid:Baz>])>
    
    Add the remaining 3 tokens (each fires the network)
    
    >>> aNode2.activate(token4.unboundCopy())
    >>> list(joinNode3.memories[LEFT_MEMORY])[0]
    <PartialInstanciation (joined on ?X): Set([<ReteToken: X->urn:uuid:Foo>, <ReteToken: X->urn:uuid:Foo,Y->urn:uuid:Baz>])>
    >>> aNode3.activate(token5.unboundCopy()))
    Propagated from <AlphaNode: (u'Z', u'urn:uuid:Prop1', u'W'). Feeds 1 beta nodes>
    (u'urn:uuid:Bar', u'urn:uuid:Prop1', u'urn:uuid:Beezle')
    <TerminalNode : CommonVariables: [] (1 in left, 1 in right memories)>.propagate(right,None,<ReteToken: Z->urn:uuid:Bar,W->urn:uuid:Beezle>)
    activating with <PartialInstanciation (joined on ?X): Set([<ReteToken: Z->urn:uuid:Bar,W->urn:uuid:Beezle>, <ReteToken: X->urn:uuid:Foo>, <ReteToken: X->urn:uuid:Foo,Y->urn:uuid:Baz>])>

    >>> aNode3.activate(token6.unboundCopy())
    >>> joinNode3
    <TerminalNode : CommonVariables: [] (1 in left, 2 in right memories)>
    >>> testHelper.firings
    2
    >>> pprint(testHelper.conflictSet)
    Set([<PartialInstanciation (joined on ?X): Set([<ReteToken: Z->urn:uuid:Bar,W->urn:uuid:Beezle>, <ReteToken: X->urn:uuid:Foo>, <ReteToken: X->urn:uuid:Foo,Y->urn:uuid:Baz>])>, <PartialInstanciation (joined on ?X): Set([<ReteToken: Z->urn:uuid:Bar,W->urn:uuid:Bundle>, <ReteToken: X->urn:uuid:Foo>, <ReteToken: X->urn:uuid:Foo,Y->urn:uuid:Baz>])>])
    """
    def __init__(self,leftNode,rightNode,aPassThru=False):
        self.instanciatingTokens = set()
        self.aPassThru = aPassThru 
        self.name = BNode()
        self.network = None
        self.consequent = Set() #List of tuples in RHS
        self.leftNode = leftNode
        self.rightNode = rightNode #The incoming right input of a BetaNode is always an AlphaNode
        self.memories = {}
        self.descendentMemory = []
        self.descendentBetaNodes = Set()
        self.leftUnlinkedNodes = Set()
        self.unlinkedMemory = None
        self.fedByBuiltin = None
        if isinstance(leftNode,BuiltInAlphaNode):
            self.fedByBuiltin = LEFT_MEMORY
            assert not isinstance(rightNode,BuiltInAlphaNode),"Both %s and %s are builtins feeding a beta node!"%(leftNode,rightNode)
            self.memories[RIGHT_MEMORY] = ReteMemory((self,RIGHT_MEMORY,leftNode.n3builtin))
        else:
            self.memories[RIGHT_MEMORY] = ReteMemory(self,RIGHT_MEMORY)
            
        assert not(self.fedByBuiltin),"No support for 'built-ins', function symbols, or non-equality tests"
            
        if isinstance(rightNode,BuiltInAlphaNode):
            self.fedByBuiltin = RIGHT_MEMORY
            assert not isinstance(leftNode,BuiltInAlphaNode),"Both %s and %s are builtins feeding a beta node!"%(leftNode,rightNode)
            self.memories[LEFT_MEMORY]  = ReteMemory(self,LEFT_MEMORY,rightNode.n3builtin)
        else:
            self.memories[LEFT_MEMORY]  = ReteMemory(self,LEFT_MEMORY)        
        if aPassThru:
            if rightNode:
                self.leftVariables = Set()
                self.rightVariables = collectVariables(self.rightNode)
                self.commonVariables = list(self.rightVariables)
            else:
                self.leftVariables = self.rightVariables = Set()
                self.commonVariables = []
        else: 
            self.leftVariables = collectVariables(self.leftNode)
            self.rightVariables = collectVariables(self.rightNode)        
            self.commonVariables = [leftVar for leftVar in self.leftVariables if leftVar in self.rightVariables]
        self.leftIndex  = {}
        self.rightIndex = {}

    def connectIncomingNodes(self,leftNode,rightNode):
        if leftNode:
            if self.leftNode and LEFT_UNLINKING:
                #candidate for left unlinking
                self.leftUnlinkedNodes.add(leftNode) 
                leftNode.unlinkedMemory = ReteMemory(self,LEFT_MEMORY)
#                print "unlinked %s from %s"%(leftNode,self)
            elif self.leftNode:            
                leftNode.descendentMemory.append(self.memories[LEFT_MEMORY])
                leftNode.descendentBetaNodes.add(self)        
        rightNode.descendentMemory.append(self.memories[RIGHT_MEMORY])
        rightNode.descendentBetaNodes.add(self)
        
    def __repr__(self):
        if self.consequent and self.fedByBuiltin:
            nodeType = 'TerminalBuiltin(%s)'%(self.memories[self._oppositeMemory(self.fedByBuiltin)].filter)
        elif self.consequent:
            nodeType = 'TerminalNode'
        elif self.fedByBuiltin:
            nodeType = 'Builtin(%s)'%(self.memories[self._oppositeMemory(self.fedByBuiltin)].filter)
        else:            
            nodeType = 'BetaNode'
        if self.unlinkedMemory is not None:
            nodeType = 'LeftUnlinked-' + nodeType
        leftLen = self.memories[LEFT_MEMORY] and len(self.memories[LEFT_MEMORY]) or 0
        return "<%s %s: CommonVariables: %s (%s in left, %s in right memories)>"%(nodeType,self.aPassThru and "(pass-thru)" or '',self.commonVariables,leftLen,len(self.memories[RIGHT_MEMORY]))

    def _activate(self,partInstOrList,debug=False):
            if debug:
                print "activating with %s"%(partInstOrList)
            if self.unlinkedMemory is not None:
                if debug:
                    print "adding %s into unlinked memory"%(partInstOrList)                
                self.unlinkedMemory.addToken(partInstOrList,debug)                
            for memory in self.descendentMemory:  
                if debug:
                    print "\t",memory.successor
                #print self,partInstOrList
                memory.addToken(partInstOrList,debug)
                if memory.successor.aPassThru or not memory.successor.checkNullActivation(memory.position):
                    if memory.position == LEFT_MEMORY:
                        memory.successor.propagate(memory.position,debug,partInstOrList)
                    else:
                        #print partInstOrList
                        memory.successor.propagate(None,debug,partInstOrList)
            
            if self.consequent:
                self.network.fireConsequent(partInstOrList,self,debug)

    def _unrollTokens(self,iterable):
        for token in iterable:
            if isinstance(token,PartialInstanciation):
                for i in token:
                    yield i
            else:
                yield token
                                             
    def _oppositeMemory(self,memoryPosition):
        if memoryPosition == LEFT_MEMORY:
            return RIGHT_MEMORY
        else:
            return LEFT_MEMORY
            
    def _checkOpposingMemory(self,memoryPosition):
        return bool(len(self.memories[self._oppositeMemory(memoryPosition)]))

    def checkNullActivation(self,source):
        """
        Checks whether this beta node is involved in a NULL activation relative to the source.
        NULL activations are where one of the opposing memories that feed
        this beta node are empty.  Takes into account built-in filters/function.
        source is the position of the 'triggering' memory (i.e., the memory that had a token added)
        """        
        oppositeMem = self.memories[self._oppositeMemory(source)]
        return not self.fedByBuiltin and not oppositeMem
            
    def propagate(self,propagationSource,debug = False,partialInst=None,wme=None):
        """
        .. 'activation' of Beta Node - checks for consistent 
        variable bindings between memory of incoming nodes ..
        Beta (join nodes) with no variables in common with both ancestors
        activate automatically upon getting a propagation 'signal'

        """
        if debug and propagationSource:
            print "%s.propagate(%s,%s,%s)"%(self,memoryPosition[propagationSource],partialInst,wme)
            print "### Left Memory ###"
            pprint(list(self.memories[LEFT_MEMORY]))
            print "###################"
            print "### Right Memory ###"
            pprint(list(self.memories[RIGHT_MEMORY]))
            print "####################"
        if self.aPassThru:
            if self.consequent:
                assert not partialInst,"%s,%s"%(partialInst,wme)
                self._activate(PartialInstanciation([wme],consistentBindings=wme.bindingDict.copy()),debug)                
                
            elif self.memories[RIGHT_MEMORY]:
                #pass on wme as an unjoined partInst
                #print self
                if wme:
                    self._activate(PartialInstanciation([wme],consistentBindings=wme.bindingDict.copy()),debug)
                elif partialInst:
                    #print "## Problem ###"
                    #print "%s.propagate(%s,%s,%s)"%(self,memoryPosition[propagationSource],partialInst,wme)
                    self._activate(partialInst,debug)                
        elif not propagationSource:
            #Beta node right activated by another beta node
            #Need to unify on common variable hash, using the bindings
            #provided by the partial instanciation that triggered the activation
            if partialInst:
                for binding in partialInst.bindings:
                    commonVals = tuple([binding[var] for var in self.commonVariables])
                    lTokens = self.memories[RIGHT_MEMORY].substitutionDict.get(commonVals,Set())
                    rTokens = self.memories[LEFT_MEMORY].substitutionDict.get(commonVals,Set())
                    joinedTokens = Set(self._unrollTokens(rTokens | lTokens))
                    if joinedTokens:                    
                        commonDict = dict([(var,list(commonVals)[self.commonVariables.index(var)]) for var in self.commonVariables])
                        newP = PartialInstanciation(joinedTokens,consistentBindings=commonDict)
                        self._activate(newP,debug)            
            
        elif propagationSource == LEFT_MEMORY:
            #Doesn't check for null left activation! - cost is mitigated by 
            #left activation, partialInst passed down
            #procedure join-node-left-activation (node: join-node, t: token)
            #     for each w in node.amem.items do
            #         if perform-join-tests (node.tests, t, w) then
            #             for each child in node.children do left-activation (child, t, w)
            # end            
            matches = Set()
            if self.fedByBuiltin:
                filter = self.memories[self._oppositeMemory(self.fedByBuiltin)].filter
                newConsistentBindings = [term for term in [filter.argument,
                                                           filter.result] 
                                                if isinstance(term,Variable) and \
                                                term not in partialInst.joinedBindings]
                partialInst.addConsistentBinding(newConsistentBindings)
                for binding in partialInst.bindings:
                    lhs = filter.argument
                    rhs = filter.result
    #                if (isinstance(lhs,Variable) and lhs not in binding) or\
    #                   (isinstance(rhs,Variable) and rhs not in binding):
        #                        if self.consequent:
    #                    print "^ \t",partialInst,self, filter
    #                    pprint(binding)
    #                    raise 
                    lhs = isinstance(lhs,Variable) and binding[lhs] or lhs
                    rhs = isinstance(rhs,Variable) and binding[rhs] or rhs
                    assert lhs is not None and rhs is not None
                    if filter.func(lhs,rhs):
                        matches.add(partialInst)
            else:
                for binding in partialInst.bindings:
                    #iterate over the binding combinations 
                    #and use the substitutionDict in the right memory to find matching WME'a
                    if debug:
                        print "\t", binding
                        
                    substitutedTerm=[]
                    commonDictKV=[]
                    for var in self.commonVariables:
                        if var not in binding:
                            continue
                        else:
                            commonDictKV.append((var,binding[var]))
                            substitutedTerm.append(binding[var])
                    rWMEs = self.memories[RIGHT_MEMORY].substitutionDict.get(tuple(substitutedTerm),
                                                                             Set())
                    commonDict = dict(commonDictKV)
                    if debug:
                        print commonDict,rWMEs, self.memories[RIGHT_MEMORY].substitutionDict.keys()
                    for rightWME in rWMEs:
                        if isinstance(rightWME,ReteToken):
                            matches.add(partialInst.newJoin(
                                rightWME,
                                ifilter(lambda x:x not in partialInst.joinedBindings,
                                    self.commonVariables)))
                            # [var for var in self.commonVariables if var not in partialInst.joinedBindings]))
                        else:
                            #Joining two Beta/Join nodes!
                            joinedTokens = list(partialInst.tokens | rightWME.tokens)
                            #print "### joining two tokens ###"
                            #pprint(joinedTokens)
                            if self.consequent:
                                for consequent in self.consequent:
                                    consVars = ifilter(lambda x:isinstance(x,Variable),consequent)
                                    # [i for i in consequent if isinstance(i,Variable)]                                
                                failed = True
                                for binding in PartialInstanciation(joinedTokens,consistentBindings=commonDict).bindings:
                                    if any(consVars,lambda x:x not in binding):# [key for key in consVars if key not in binding]:
                                        continue
                                    else:
                                        failed = False                                                                    
                                if not failed:                                        
                                    newP = PartialInstanciation(joinedTokens,consistentBindings=commonDict)
                                    matches.add(newP)
                            else:
                                newP = PartialInstanciation(joinedTokens,consistentBindings=commonDict)
                                matches.add(newP)
                                
            for pInst in matches:
                self._activate(pInst,debug)                    
        else:            
            #right activation, partialInst & wme passed down
            #procedure join-node-right-activation (node: join-node, w: WME)
            #    for each t in node.parent.items do {"parent" is the beta memory node}
            #        if perform-join-tests (node.tests, t, w) then
            #            for each child in node.children do left-activation (child, t, w)
            #end
            #pprint(self.memories[self._oppositeMemory(propagationSource)])
            matches = Set()
            lPartInsts = self.memories[LEFT_MEMORY].substitutionDict.get(tuple([wme.bindingDict[var] for var in self.commonVariables]))
            if lPartInsts:
                for partialInst in lPartInsts:
                    if not isinstance(partialInst,PartialInstanciation):
                        singleToken = PartialInstanciation([partialInst],consistentBindings=partialInst.bindingDict.copy())
                        matches.add(singleToken)
                    else:
                        assert isinstance(partialInst,PartialInstanciation),repr(partialInst)
                        matches.add(partialInst.newJoin(
                                        wme,
                                        ifilter(lambda x:x not in partialInst.joinedBindings,
                                                self.commonVariables)))
                                        # [var for var in self.commonVariables if var not in partialInst.joinedBindings]))
            for pInst in matches:
                self._activate(pInst,debug)  

TEST_NS = Namespace('http://example.com/text1/')

def PopulateTokenFromANode(aNode,bindings):
    #print aNode, bindings
    termList = [isinstance(term,Variable) and bindings[term] or term
                    for term in aNode.triplePattern]
    token = ReteToken(tuple(termList))
    token.bindVariables(aNode)
    return token

class PartialInstanciationTests(unittest.TestCase):
    def testConsistentBinding(self):
        allBindings = {}
        allBindings.update(self.joinedBindings)
        allBindings.update(self.unJoinedBindings)
        aNodes = [self.aNode1,
                  self.aNode2,
                  self.aNode5,
                  self.aNode6,
                  self.aNode7,
                  self.aNode8,
                  self.aNode9,
                  self.aNode10,
                  self.aNode11]
        pToken = PartialInstanciation(
                      tokens = [PopulateTokenFromANode(aNode,
                                                       allBindings) 
                                  for aNode in aNodes],
                      consistentBindings = self.joinedBindings)
        #print pToken
        pToken.addConsistentBinding(self.unJoinedBindings.keys())
        #print pToken.joinedBindings
        for binding in pToken.bindings:
            for key in self.unJoinedBindings:
                self.failUnless(key in binding, "Missing key %s from %s"%(key,binding))
                        
    def setUp(self):
        self.aNode1 = AlphaNode((Variable('HOSP'),
                                 TEST_NS.contains,
                                 Variable('HOSP_START_DATE')))                
        self.aNode2 = AlphaNode((Variable('HOSP'),
                                 RDF.type,
                                 TEST_NS.Hospitalization))                
        self.aNode5 = AlphaNode((Variable('HOSP_START_DATE'),
                                 TEST_NS.dateTimeMin,
                                 Variable('ENCOUNTER_START')))                
        self.aNode6 = AlphaNode((Variable('HOSP_STOP_DATE'),
                                 RDF.type,
                                 TEST_NS.EventStopDate))                
        self.aNode7 = AlphaNode((Variable('HOSP_STOP_DATE'),
                                 TEST_NS.dateTimeMax,
                                 Variable('ENCOUNTER_STOP')))                
        self.aNode8 = AlphaNode((Variable('EVT_DATE'),
                                 RDF.type,
                                 TEST_NS.EventStartDate))                
        self.aNode9 = AlphaNode((Variable('EVT_DATE'),
                                 TEST_NS.dateTimeMin,
                                 Variable('EVT_START_MIN')))                
        self.aNode10 =AlphaNode((Variable('EVT'),
                                 TEST_NS.contains,
                                 Variable('EVT_DATE')))                
        self.aNode11 =AlphaNode((Variable('EVT'),
                                 RDF.type,
                                 Variable('EVT_KIND')))

        self.joinedBindings = {Variable('HOSP_START_DATE'):
                               BNode(),
                               Variable('HOSP_STOP_DATE'):
                               BNode(),
                               Variable('HOSP'):
                               BNode()}
        self.unJoinedBindings = {Variable('EVT'):
                                 BNode(),
                                 Variable('EVT_DATE'):
                                 BNode(),
                                 Variable('EVT_KIND'):
                                 TEST_NS.ICUStay}
        for dtVariable in [Variable('ENCOUNTER_START'),
                           Variable('ENCOUNTER_STOP'),
                           Variable('EVT_START_MIN')]:
            self.unJoinedBindings[dtVariable]=Literal("2007-02-14T10:00:00",
                                                      datatype=_XSD_NS.dateTime)
                                  
def test():
#    import doctest
#    doctest.testmod()
    suite = unittest.makeSuite(PartialInstanciationTests)
    unittest.TextTestRunner(verbosity=5).run(suite)    

if __name__ == '__main__':
    test()


"""
"""

class Node:
    """
    A node in a Rete network.  Behavior between Alpha and Beta (Join) nodes
    """
    def connectToBetaNode(self,betaNode,position):
        from BetaNode import BetaNode, LEFT_MEMORY, RIGHT_MEMORY, PartialInstanciation
        self.descendentMemory.append(betaNode.memories[position])
        self.descendentBetaNodes.add(betaNode)        